import copy
from unittest.mock import MagicMock, ANY

import pytest

from refinitiv.data._core.session import SessionType, DesktopSession
from refinitiv.data._content_type import ContentType
from refinitiv.data.content import pricing
from refinitiv.data.content._universe_streams import (
    _UniverseStreams,
    UniverseStreamFacade,
)
from refinitiv.data.delivery._stream import StreamState, StreamEvent
from tests.unit.conftest import StubSession
from concurrent.futures import ThreadPoolExecutor, wait


def test_get_item_error():
    # given
    session = StubSession()
    universe = ["RIC"]

    # when
    stream = _UniverseStreams(
        content_type=ContentType.NONE, universe=universe, session=session
    )

    # then
    assert stream["__mock__"] == {}


def test_universe_str():
    # given
    session = StubSession()
    universe = "RIC"

    try:
        # when
        _UniverseStreams(
            content_type=ContentType.NONE, universe=universe, session=session
        )
    except Exception as e:
        assert False, e


def test_stream__repr__(streams):
    # when
    testing_value = repr(streams)

    # then
    assert testing_value is not None


def test_stream__iter__(streams):
    # when
    iterator = iter(streams)

    # then
    assert next(iterator) is not None


def test_stream__getitem__(streams):
    # when
    streaming_price = streams["item"]

    # then
    assert streaming_price == {}


def test_stream__getitem__wrapper(streams):
    # when
    streams = streams["name"]

    # then
    assert isinstance(streams, UniverseStreamFacade)


def test_stream__len__(streams):
    # when
    testing_value = len(streams)

    # then
    assert testing_value == 1


def test_call_streams_close_for_unopened_streams(streams):
    # when
    state = streams.close()

    # then
    assert state is StreamState.Unopened, state


def test_stream_on_update(streams):
    # given
    callback_called = False

    def on_update(owner, name, fields):
        nonlocal callback_called
        callback_called = True

    streams.on_update(on_update)

    # when
    streams["name"]._stream._on_stream_update({}, {"Fields": {}})

    # then
    assert callback_called is True


def test_stream_on_status(streams):
    # given
    callback_called = False

    def on_status(owner, name, fields):
        nonlocal callback_called
        callback_called = True

    streams.on_status(on_status)

    # when
    streams["name"]._stream._on_stream_status({}, {"State": {}, "Fields": {}})

    # then
    assert callback_called is True


def test_stream_on_complete(streams):
    # given
    callback_called = False

    def on_complete(owner):
        nonlocal callback_called
        callback_called = True

    streams.on_complete(on_complete)

    # when
    streams["name"]._stream._on_stream_complete({})

    # then
    assert callback_called is True


def test_stream_on_error(streams):
    # given
    callback_called = False

    def on_error(owner, name, fields):
        nonlocal callback_called
        callback_called = True

    streams.on_error(on_error)

    # when
    streams["name"]._stream._on_stream_error({}, {})

    # then
    assert callback_called is True


def test_stream_keys(streams):
    keys = streams.keys()

    assert keys is not None


def test_stream_values(streams):
    values = streams.values()

    assert values is not None


def test_stream_items(streams):
    items = streams.items()

    assert items is not None


def test_stream___iter___raise_error(streams):
    iterator = iter(streams)

    next(iterator)

    with pytest.raises(StopIteration):
        next(iterator)


def test_stream___iter___():
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        content_type=ContentType.NONE,
        universe="universe",
        fields=["field"],
        session=session,
    )

    iterator = iter(stream)
    testing_value = next(iterator)

    assert testing_value is not None


def test_stream___len__(streams):
    length = len(streams)

    assert length == 1


def test_stream___str__(streams):
    s = str(streams)

    assert s is not None


def test__do_on_stream_update():
    # given
    data = [
        {
            "Fields": {"DATETIME": "2020-12-31T11:55:59.000Z", "TRDPRC_1": 123.456},
            "ID": 2,
            "Key": {"Name": "S)MySymbol.UUID", "Service": "CUSTOM_BASKETS"},
            "Type": "Update",
            "UpdateType": "Quote",
        }
    ]
    test_result = copy.copy(data[0])

    # when
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        content_type=ContentType.NONE, universe="", session=session
    )
    stream._record = data
    mock_stream = MagicMock()
    mock_stream.name = "Stream"
    result = stream._do_on_stream_update(mock_stream, *data)

    # then
    assert result[1] == test_result
    assert stream._record == data


def test__do_on_stream_refresh():
    # given
    data = [
        {
            "Fields": {"DATETIME": "2020-12-31T11:55:59.000Z", "TRDPRC_1": 123.456},
            "ID": 2,
            "Key": {"Name": "S)MySymbol.UUID", "Service": "CUSTOM_BASKETS"},
            "Type": "Update",
            "UpdateType": "Quote",
        }
    ]
    test_result = copy.copy(data[0])

    # when
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        content_type=ContentType.NONE, universe="", session=session
    )
    stream._record = data
    mock_stream = MagicMock()
    mock_stream.name = "Stream"
    result = stream._do_on_stream_refresh(mock_stream.MagicMock(), *data)

    # then
    assert result[1] == test_result
    assert stream._record == data


def test__add_instruments_when_close():
    test_data = ["VOD.L", "EUR=", "IBM"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(ContentType.NONE, universe="", session=session)
    stream.add_instruments(test_data)
    for i in test_data:
        assert i in stream._universe
    assert not stream.__dict__.get("_stream_by_name")


def test__remove_instruments_when_close():
    test_data = ["VOD.L", "IBM"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE, universe=["VOD.L", "EUR=", "IBM"], session=session
    )
    stream.remove_instruments(test_data)

    for i in test_data:
        assert i not in stream._universe
    assert not stream.__dict__.get("_stream_by_name")
    assert "EUR=" in stream._universe


def test__add_instruments_when_open():
    test_data = ["VOD.L", "EUR=", "IBM"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(ContentType.NONE, universe="", session=session)

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._create_stream_by_name = lambda name: MagicMock()

    stream.add_instruments(test_data)
    for i in test_data:
        assert i in stream._stream_by_name
        assert i in stream._universe
    stream.close()


def test__remove_instruments_when_open():
    test_data = ["VOD.L", "IBM"]

    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE, universe=["VOD.L", "EUR=", "IBM"], session=session
    )
    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._create_stream_by_name = lambda name: MagicMock()

    stream.remove_instruments(test_data)
    for i in test_data:
        assert i not in stream._stream_by_name
        assert i not in stream._universe
    assert "EUR=" in stream._stream_by_name
    assert "EUR=" in stream._universe


def test__remove_instruments_when_universe_is_empty():
    test_data = ["VOD.L", "IBM"]
    call_counter = 0

    def _error(_):
        nonlocal call_counter
        call_counter += 1

    session = StubSession(is_open=True)
    stream = _UniverseStreams(ContentType.NONE, universe=[], session=session)
    stream._error = _error
    stream.remove_instruments(test_data)
    assert call_counter == 1


def test__remove_instruments_when_name_not_in_universe():
    test_data = ["VOD.L", "IBM"]
    call_counter = 0

    def _error(_):
        nonlocal call_counter
        call_counter += 1

    session = StubSession(is_open=True)
    stream = _UniverseStreams(ContentType.NONE, universe=["EUR="], session=session)
    stream._error = _error
    stream.remove_instruments(test_data)
    assert call_counter == 2


def test__only_one_complete():
    def on_complete(*args, **kwargs):
        pass

    test_data = ["VOD.L", "EUR=", "IBM"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(ContentType.NONE, universe="", session=session)
    stream.on_complete(on_complete)

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._create_stream_by_name = lambda name: MagicMock()

    stream.add_instruments(test_data)
    for i in test_data:
        stream._stream_by_name[i].off.assert_called_with(StreamEvent.COMPLETE, ANY)


def test__add_fields_when_close():
    test_data = ["BID", "ASK"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1"],
        session=session,
    )
    stream.add_fields(test_data)

    for i in test_data:
        assert i in stream.fields
    assert not stream.__dict__.get("_stream_by_name")


def test__add_fields_when_open():
    test_data = ["BID", "ASK"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...

    stream.add_fields(test_data)
    for i in test_data:
        assert i in stream.fields
    stream.close()


def test__remove_fields_when_close():
    test_data = ["BID", "ASK"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1", "BID", "ASK"],
        session=session,
    )
    stream.remove_fields(test_data)

    for i in test_data:
        assert i not in stream.fields
    assert not stream.__dict__.get("_stream_by_name")


def test__remove_fields_when_open():
    test_data = ["BID", "ASK"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1", "BID", "ASK"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...

    stream.remove_fields(test_data)
    for i in test_data:
        assert i not in stream.fields
    stream.close()


def make_error(expected_msg):
    def _error(msg):
        assert msg == expected_msg

    return make_error


def test_remove_not_exists_field():
    test_data = ["INVAL"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1", "BID", "ASK"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...
    stream._error = make_error("{'INVAL'} not in fields list")
    stream.remove_fields(test_data)

    stream.close()


def test_add_exists_field():
    test_data = ["TRDPRC_1"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1", "BID", "ASK"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...
    stream._error = make_error("{'TRDPRC_1'} already in fields list")
    stream.add_fields(test_data)

    stream.close()


def test_remove_not_exists_item():
    test_data = ["INVAL"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1", "BID", "ASK"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...
    stream._error = make_error("{'INVAL'} not in universe list")
    stream.remove_instruments(test_data)

    stream.close()


def test_add_exists_item():
    test_data = ["VOD.L"]
    session = StubSession(is_open=True)
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1", "BID", "ASK"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...
    stream._error = make_error("{'TRDPRC_1'} already in universe list")
    stream.add_instruments(test_data)

    stream.close()


class StubDesktop(DesktopSession, StubSession):
    def __init__(self, *args, **kwargs):
        StubSession.__init__(self, *args, **kwargs)


def test__add_fields_when_open_desktop_session():
    test_data = ["BID", "ASK"]
    session = StubDesktop(is_open=True)
    session.type = SessionType.DESKTOP
    stream = _UniverseStreams(
        ContentType.NONE,
        universe=["VOD.L", "EUR=", "IBM"],
        fields=["TRDPRC_1"],
        session=session,
    )

    # open
    _ = stream._stream_by_name
    stream._state = StreamState.Opened
    stream._reopen_all_streams = lambda: ...
    stream._stream_by_name = {"a": MagicMock(), "B": MagicMock()}

    stream.add_fields(test_data)
    for i in test_data:
        assert i in stream.fields
        for stream_mock in stream._stream_by_name.values():
            stream_mock.off.assert_called()
            stream_mock.close.assert_called()
            stream_mock.open.assert_called()
    stream.close()


def test_add_items_and_remove_items_do_not_work_at_the_same_time():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(universe=["VOD.L", "IBM.N"])
    stream = definition.get_stream(session)

    def add_items():
        stream.add_items(["EUR="])
        # then
        assert "EUR=" in stream._universe

    def remove_items():
        stream.remove_items(["EUR="])
        # then
        assert "EUR=" not in stream._universe

    # when
    with ThreadPoolExecutor() as ex:
        futures = []
        for _ in range(100):
            futures.extend([ex.submit(add_items), ex.submit(remove_items)])
        wait(futures)


def test_add_fields_and_remove_fields_do_not_work_at_the_same_time():
    # given
    session = StubSession(is_open=True)
    definition = pricing.Definition(universe=["VOD.L"], fields=["BID", "ASK"])
    stream = definition.get_stream(session)

    def add_fields():
        stream.add_fields(["TRDPRC_1"])
        # then
        assert "TRDPRC_1" in stream._universe

    def remove_fields():
        stream.remove_fields(["TRDPRC_1"])
        # then
        assert "TRDPRC_1" not in stream._universe

    # when
    with ThreadPoolExecutor() as ex:
        futures = []
        for _ in range(100):
            futures.extend([ex.submit(add_fields), ex.submit(remove_fields)])
        wait(futures)



from typing import List, TYPE_CHECKING

from .._content_data_validator import ContentDataValidator
from .._request_factory import DatesAndCalendarsResponseFactory
from ..holidays._holidays_data_provider import Holiday
from ...._content_data import Data
from ...._content_data_provider import ContentDataProvider
from ...._df_builder import build_dates_calendars_df
from .....content.ipa._content_provider import DatesAndCalendarsRequestFactory
from .....delivery._data._data_provider import ValidatorContainer

if TYPE_CHECKING:
    from .....delivery._data._data_provider import ParsedData


class WorkingDay:
    def __init__(
        self,
        is_weekend: bool,
        is_working_day: bool,
        tag: str = "",
        holidays: list = None,
    ):
        self._is_weekend = is_weekend
        self._response_holidays_items = holidays or []
        self._is_working_day = is_working_day
        self._tag = tag
        self._holidays = []

    @property
    def is_weekend(self):
        return self._is_weekend

    @property
    def is_working_day(self):
        return self._is_working_day

    @property
    def tag(self):
        return self._tag

    @property
    def holidays(self) -> List[Holiday]:
        if self._holidays:
            return self._holidays

        for holiday in self._response_holidays_items:
            holiday_ = Holiday(holiday=holiday, tag=self.tag)
            self._holidays.append(holiday_)
        return self._holidays


class IsWorkingDay(Data):
    def __init__(self, raw: dict, tag: str = "", **kwargs):
        super().__init__(raw, **kwargs)
        self._day = WorkingDay(
            is_weekend=raw[0]["isWeekEnd"],
            is_working_day=raw[0]["isWorkingDay"],
            tag=tag,
            holidays=raw[0].get("holidays", []),
        )

    @property
    def day(self):
        return self._day


class IsWorkingDays(Data):
    _is_working_days_ = None

    @property
    def _is_working_days(self):
        if self._is_working_days_ is None:
            self._is_working_days_ = [
                WorkingDay(
                    raw_item["isWeekEnd"],
                    raw_item.get("isWorkingDay"),
                    tag=raw_item.get("tag"),
                    holidays=raw_item.get("holidays"),
                )
                for raw_item in self.raw
                if not raw_item.get("error")
            ]

        return self._is_working_days_

    @property
    def days(self):
        return self._is_working_days

    def __getitem__(self, item: int):
        return self._is_working_days[item]


class IsWorkingDayResponseFactory(DatesAndCalendarsResponseFactory):
    def create_data_success(self, parsed_data: "ParsedData", **kwargs):
        raw = self.get_raw(parsed_data)

        if len(raw) > 1:
            data = IsWorkingDays(raw, build_dates_calendars_df, **kwargs)

        else:
            data = IsWorkingDay(
                raw, raw[0].get("tag"), dfbuilder=build_dates_calendars_df, **kwargs
            )

        return data


is_working_day_data_provider = ContentDataProvider(
    request=DatesAndCalendarsRequestFactory(),
    response=IsWorkingDayResponseFactory(),
    validator=ValidatorContainer(content_validator=ContentDataValidator()),
)



import json

import allure
import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.errors import RDError
from tests.integration.content.ipa.contracts.conftest import option_definition_01
from tests.integration.content.ipa.contracts.option.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
    option_definition_02,
    option_universe,
)
from tests.integration.helpers import (
    get_async_response_from_definition,
    get_async_response_from_definitions,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - Option")
@allure.feature("Content object - Option")
@allure.severity(allure.severity_level.CRITICAL)
class TestOption:
    @allure.title("Create an option definition object with params and get data")
    @pytest.mark.caseid("36647914")
    def test_create_option_with_valid_params(self, open_session):
        response = option_definition_01().get_data()
        check_http_status_is_success_and_df_value_not_empty(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create an Option definition without any parameters")
    @pytest.mark.caseid("C36647917")
    def test_create_option_without_any_params(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.option.Definition().get_data(session=open_session)
        assert (
            str(error.value)
            == "Error code 400 | The underlyingType cannot be found, in the request, from the expected path instrumentDefinition.underlyingType for 'instrumentType' Option"
        )

    @allure.title("Create an Option definition with empty pricing parameters")
    @pytest.mark.caseid("36647981")
    def test_create_option_with_empty_pricing_parameters(self, open_session):
        response = rdf.option.Definition(
            underlying_type=rdf.option.UnderlyingType.FX,
            strike=265,
            underlying_definition=rdf.option.FxUnderlyingDefinition("AUDUSD"),
            notional_ccy="AUD",
            tenor="5M",
            pricing_parameters=rdf.option.PricingParameters(),
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a Option definition request asynchronously")
    @pytest.mark.caseid("36678381")
    async def test_create_async_option_definition(self, open_session_async):
        response = await get_async_response_from_definition(option_definition_01())

        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create multiple async Option definitions")
    @pytest.mark.caseid("36678382")
    async def test_create_multiple_async_option_definitions(self, open_session_async):
        response_01, response_02 = await get_async_response_from_definitions(
            option_definition_01(), option_definition_02()
        )

        check_http_status_is_success_and_df_value_not_empty(response_01)

    @allure.title("Create an Option stream")
    @pytest.mark.caseid("36678383")
    def test_create_option_stream(self, open_session):
        stream = option_definition_01().get_stream()
        stream.open()

        check_stream_state_and_df_from_stream(stream)

    @allure.title("Create an option definition with extended parameters")
    @pytest.mark.caseid("C37435263")
    def test_get_option_with_extended_params(self, open_session):
        response = rdf.option.Definition(
            underlying_type=rdf.option.UnderlyingType.FX,
            strike=290,
            underlying_definition=rdf.option.FxUnderlyingDefinition("EUR"),
            tenor="7M",
            pricing_parameters=rdf.option.PricingParameters(
                pricing_model_type=rdf.option.PricingModelType.BLACK_SCHOLES,
                fx_spot_object=rdf.option.BidAskMid(
                    bid=0.7387,
                    ask=0.7387,
                    mid=0.7387,
                ),
            ),
            extended_params={
                "instrumentDefinition": {
                    "InstrumentTag": "my tag",
                    "tenor": "5M",
                    "notionalCcy": "AUD",
                    "underlyingDefinition": {"fxCrossCode": "AUDUSD"},
                    "underlyingType": "Fx",
                    "strike": 265,
                },
                "pricingParameters": {
                    "fxSpotObject": {"bid": 0.7387, "ask": 0.7387, "mid": 0.7387},
                    "priceSide": "Mid",
                    "pricingModelType": "BlackScholes",
                    "valuationDate": "2018-08-06",
                },
            },
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == option_universe
        ), f"Extended params are applied incorrectly"
        assert response.data.df.InstrumentTag[0] == "my tag"



# coding: utf8

__all__ = ["MaturityFilter"]

from ..._object_definition import ObjectDefinition


class MaturityFilter(ObjectDefinition):
    def __init__(
        self,
        max_maturity=None,
        min_maturity=None,
        min_of_median_nb_of_strikes_percent=None,
    ):
        super().__init__()
        self.max_maturity = max_maturity
        self.min_maturity = min_maturity
        self.min_of_median_nb_of_strikes_percent = min_of_median_nb_of_strikes_percent

    @property
    def max_maturity(self):
        """
        Max Maturity to consider in the filtering. (expressed in tenor)
        :return: str
        """
        return self._get_parameter("maxMaturity")

    @max_maturity.setter
    def max_maturity(self, value):
        self._set_parameter("maxMaturity", value)

    @property
    def min_maturity(self):
        """
        Min Maturity to consider in the filtering. (expressed in tenor)
        Default value: 7D
        :return: str
        """
        return self._get_parameter("minMaturity")

    @min_maturity.setter
    def min_maturity(self, value):
        self._set_parameter("minMaturity", value)

    @property
    def min_of_median_nb_of_strikes_percent(self):
        """
        Remove maturities whose number of strikes is less than MinOfMedianNbOfStrikesPercent of the Median number of Strikes.
        :return: float
        """
        return self._get_parameter("minOfMedianNbOfStrikesPercent")

    @min_of_median_nb_of_strikes_percent.setter
    def min_of_median_nb_of_strikes_percent(self, value):
        self._set_parameter("minOfMedianNbOfStrikesPercent", value)



from datetime import datetime, timedelta
from typing import List, Optional, Union, TYPE_CHECKING

from numpy import iterable

from ._count_periods_data_provider import CountedPeriod, CountedPeriods
from .._base_request_items import StartEndDateBase
from ..._enums import DayCountBasis, PeriodType
from ...._content_provider_layer import ContentUsageLoggerMixin
from ....._content_type import ContentType
from ....._tools import create_repr, try_copy_to_list
from .....delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ....._types import ExtendedParams, OptDateTime, OptStrStrs


class CountPeriodsRequestItem(StartEndDateBase):
    def __init__(
        self,
        tag: Optional[str],
        start_date: Union[str, datetime, timedelta],
        end_date: Union[str, datetime, timedelta],
        period_type: Optional[PeriodType],
        calendars: Optional[List[str]],
        currencies: Optional[List[str]],
        day_count_basis,
    ):
        super().__init__(start_date, end_date)
        self.tag = tag
        self.calendars = calendars
        self.period_type = period_type
        self.currencies = currencies
        self.day_count_basis = day_count_basis

    @property
    def tag(self):
        """
        :return: str
        """
        return self._get_parameter("tag")

    @tag.setter
    def tag(self, value):
        self._set_parameter("tag", value)

    @property
    def calendars(self):
        """
        :return: list
        """
        return self._get_parameter("calendars")

    @calendars.setter
    def calendars(self, value):
        """
        :return: list
        """
        self._set_parameter("calendars", value)

    @property
    def period_type(self):
        """
        :return: PeriodType
        """
        return self._get_enum_parameter(PeriodType, "periodType")

    @period_type.setter
    def period_type(self, value):
        self._set_enum_parameter(PeriodType, "periodType", value)

    @property
    def currencies(self):
        """
        :return: list
        """
        return self._get_parameter("currencies")

    @currencies.setter
    def currencies(self, value):
        self._set_parameter("currencies", value)

    @property
    def day_count_basis(self):
        """
        :return: DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "dayCountBasis")

    @day_count_basis.setter
    def day_count_basis(self, value):
        self._set_enum_parameter(DayCountBasis, "dayCountBasis", value)


class Definition(
    ContentUsageLoggerMixin[BaseResponse[CountedPeriod]],
    DataProviderLayer[BaseResponse[CountedPeriod]],
):
    """
    Count periods definition object

    Parameters
    ----------
        start_date: str or datetime or timedelta
            Start date of calculation.
        end_date: str or datetime or timedelta
            End date of calculation.
        tag: str, optional
            Reference tag to map particular response in payload.
        period_type : PeriodType or str, optional
            The method we chose to count the period of time based on value from PeriodType enumeration.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.
        day_count_basis: DayCountBasis or str, optional
            Day count basis value from DayCountBasis enumeration.
        extended_params : dict, optional
            If necessary other parameters.

    Methods
    -------
    get_data(session=None, on_response=None, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> import datetime
     >>> from refinitiv.data.content.ipa.dates_and_calendars import count_periods
     >>> definition = count_periods.Definition(
     ...   tag="my request",
     ...   start_date=datetime.timedelta(-11),
     ...   end_date=datetime.timedelta(-3),
     ...   period_type=count_periods.PeriodType.WORKING_DAY,
     ...   calendars=["EMU"],
     ...   currencies=["EUR"],
     ...   day_count_basis=count_periods.DayCountBasis.DCB_30_360
     ... )
     >>> response = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    _USAGE_CLS_NAME = "IPA.DatesAndCalendars.CountPeriodsDefinition"

    def __init__(
        self,
        start_date: "OptDateTime" = None,
        end_date: "OptDateTime" = None,
        tag: Optional[str] = None,
        period_type: Optional[Union[PeriodType, str]] = None,
        calendars: "OptStrStrs" = None,
        currencies: "OptStrStrs" = None,
        day_count_basis: Optional[Union[DayCountBasis, str]] = None,
        extended_params: "ExtendedParams" = None,
    ):
        calendars = try_copy_to_list(calendars)
        currencies = try_copy_to_list(currencies)
        self.extended_params = extended_params

        self.request_item = CountPeriodsRequestItem(
            tag=tag,
            start_date=start_date,
            end_date=end_date,
            period_type=period_type,
            calendars=calendars,
            currencies=currencies,
            day_count_basis=day_count_basis,
        )

        super().__init__(
            data_type=ContentType.DATES_AND_CALENDARS_COUNT_PERIODS,
            universe=[self.request_item],
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)


DefnDefns = Union[List[Definition], Definition]


class Definitions(
    ContentUsageLoggerMixin[BaseResponse[CountedPeriods]],
    DataProviderLayer[BaseResponse[CountedPeriods]],
):
    """
    Count periods definitions object

    Parameters
    ----------
    universe: Definition or list of Definition objects
        List of initialized Definition objects to retrieve data.

    Methods
    -------
    get_data(session=None, on_response=None, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> import datetime
     >>> from refinitiv.data.content.ipa.dates_and_calendars import count_periods
     >>> first_definition = count_periods.Definition(
     ...   tag="my request",
     ...   start_date=datetime.timedelta(-11),
     ...   end_date=datetime.timedelta(-3),
     ...   period_type=count_periods.PeriodType.WORKING_DAY,
     ...   calendars=["EMU"],
     ...   currencies=["EUR"],
     ...   day_count_basis=count_periods.DayCountBasis.DCB_30_360
     ... )
     ...
     >>> second_definition = count_periods.Definition(
     ...   tag="my second request",
     ...   start_date=datetime.timedelta(-15),
     ...   end_date=datetime.timedelta(-10),
     ...   period_type=count_periods.PeriodType.NON_WORKING_DAY,
     ...   calendars=["EMU"],
     ...   currencies=["EUR"],
     ...   day_count_basis=count_periods.DayCountBasis.DCB_30_360
     ... )
     >>> response = count_periods.Definitions([first_definition, second_definition]).get_data()

    Using get_data_async
     >>> import asyncio
     >>> task = count_periods.Definitions([first_definition, second_definition]).get_data_async()
     >>> response = asyncio.run(task)
    """

    _USAGE_CLS_NAME = "IPA.DatesAndCalendars.CountPeriodsDefinitions"

    def __init__(self, universe: "DefnDefns"):
        universe = try_copy_to_list(universe)
        if not iterable(universe):
            universe = [universe]

        request_items = []
        extended_params = []
        for item in universe:
            request_items.append(item.request_item)
            extended_params.append(item.extended_params)

        super().__init__(
            data_type=ContentType.DATES_AND_CALENDARS_COUNT_PERIODS,
            universe=request_items,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



import pytest

from refinitiv.data._tools._common import (
    args_parser,
    is_all_str,
    hp_universe_parser,
)


def test_args_parser_with_str():
    test_str = "EUR"

    result = args_parser(test_str)

    assert result == [test_str], f"String is not wrapped in list, result: {test_str}"


def test_args_parser_with_expect_none():
    input_val = None
    result = args_parser(input_val, expect_none=True)

    assert result is None, f"initial input was changed, result:{result}"


def test_args_parser_with_list():
    input_val = ["1", "2", "3"]
    result = args_parser(input_val)

    assert result == input_val, f"initial input was changed, result:{result}"


@pytest.mark.parametrize(
    "input_val",
    [
        123,
        ["1", 2, 3],
        (1, 2, 3),
        {"1": 1, "2": 3},
        None,
    ],
)
def test_args_parser_with_invalid_inputs(input_val):
    with pytest.raises(ValueError):
        args_parser(input_val)


@pytest.mark.parametrize(
    "input_val, expected",
    [
        ([1, "3", 2], False),
        (["1", "3", "2"], True),
    ],
)
def test_only_str_in(input_val, expected):
    assert is_all_str(input_val) == expected


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (["", " ", "boo"], ["boo"]),
        (["foo", "boo", "foo"], ["foo", "boo", "foo"]),
    ],
)
def test_hp_universe_parser(input_value, expected_value):
    actual = hp_universe_parser.get_list(input_value)

    assert actual == expected_value


def test_hp_universe_parser_with_empty_list():
    with pytest.raises(ValueError):
        hp_universe_parser.get_list([])



import asyncio
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, wait
from functools import partial
from typing import List, TYPE_CHECKING

from ._content_data import Data
from ._content_data_provider import ContentDataProvider
from ._entire_data_provider import get_entire_data_provider
from ._historical_join_responses import historical_join_responses
from ._intervals import DayIntervalType, get_day_interval_type
from .._errors import RDError
from .._tools import fields_arg_parser
from ..delivery._data._data_provider import DataProvider


if TYPE_CHECKING:
    from .._content_type import ContentType
    from ._df_build_type import DFBuildType
    from .._types import Strings
    from ..delivery._data._data_provider import Response
    from ._historical_df_builder import HistoricalBuilder


def copy_fields(fields: List[str]) -> List[str]:
    if fields is None:
        return []

    if not isinstance(fields, (list, str)):
        raise AttributeError(f"fields not support type {type(fields)}")
    fields = fields_arg_parser.get_list(fields)

    return fields[:]


def get_first_success_response(responses: List["Response"]) -> "Response":
    successful = (response for response in responses if response.is_success)
    first_successful = next(successful, None)
    return first_successful


def validate_responses(responses: List["Response"]):
    response = get_first_success_response(responses)

    if response is None:
        error_message = "ERROR: No successful response.\n"

        error_codes = set()

        for response in responses:
            if response.errors:
                error = response.errors[0]

                if error.code not in error_codes:
                    error_codes.add(error.code)
                    sub_error_message = error.message

                    if "." in error.message:
                        sub_error_message, _ = error.message.split(".", maxsplit=1)

                    error_message += f"({error.code}, {sub_error_message}), "

        error_message = error_message[:-2]
        error = RDError(1, f"No data to return, please check errors: {error_message}")
        error.response = responses
        raise error


class HistoricalDataProvider(ContentDataProvider):
    @abstractmethod
    def _get_axis_name(self, interval, **kwargs) -> str:
        # for override
        pass

    def get_data(self, *args, **kwargs) -> "Response":
        universe: List[str] = kwargs.pop("universe", [])
        entire_data_provider = get_entire_data_provider(kwargs.get("__content_type__"))

        with ThreadPoolExecutor(thread_name_prefix="HistoricalRequestThread") as ex:
            futures = []
            for inst_name in universe:
                fut = ex.submit(
                    entire_data_provider.get_data,
                    partial(super().get_data, *args),
                    universe=inst_name,
                    **kwargs,
                )
                futures.append(fut)

            wait(futures)

            responses = []
            for fut in futures:
                exception = fut.exception()

                if exception:
                    raise exception

                responses.append(fut.result())

        validate_responses(responses)

        return self._process_responses(
            responses,
            universe,
            copy_fields(kwargs.get("fields")),
            kwargs.get("interval"),
            kwargs.get("__content_type__"),
            kwargs.get("__dfbuild_type__"),
        )

    async def get_data_async(self, *args, **kwargs) -> "Response":
        universe: List[str] = kwargs.pop("universe", [])
        entire_data_provider = get_entire_data_provider(kwargs.get("__content_type__"))

        tasks = []
        for inst_name in universe:
            tasks.append(
                entire_data_provider.get_data_async(
                    partial(super().get_data_async, *args), universe=inst_name, **kwargs
                )
            )

        responses = await asyncio.gather(*tasks)
        if len(responses) == 1 and get_first_success_response(responses) is None:
            return responses.pop()

        return self._process_responses(
            responses,
            universe,
            copy_fields(kwargs.get("fields")),
            kwargs.get("interval"),
            kwargs.get("__content_type__"),
            kwargs.get("__dfbuild_type__"),
        )

    def _process_responses(
        self,
        responses: List["Response"],
        universe: "Strings",
        fields: "Strings",
        interval,
        content_type: "ContentType",
        dfbuild_type: "DFBuildType",
    ) -> "Response":
        df_builder: "HistoricalBuilder" = self.response.get_dfbuilder(
            content_type, dfbuild_type
        )

        if len(responses) == 1:
            raw = responses[0].data.raw
            data = Data(
                raw,
                dfbuilder=partial(
                    df_builder.build_one,
                    fields=fields,
                    axis_name=self._get_axis_name(interval),
                ),
            )
        else:
            raws = [response.data.raw for response in responses]
            data = Data(
                raws,
                dfbuilder=partial(
                    df_builder.build,
                    universe=universe,
                    fields=fields,
                    axis_name=self._get_axis_name(interval),
                ),
            )

        response = historical_join_responses(responses, data)
        return response


field_timestamp_by_day_interval_type = {
    DayIntervalType.INTER: "DATE",
    DayIntervalType.INTRA: "DATE_TIME",
}

axis_by_day_interval_type = {
    DayIntervalType.INTRA: "Timestamp",
    DayIntervalType.INTER: "Date",
}


def get_fields_events(fields, **kwargs):
    fields = fields_arg_parser.get_list(fields)
    result = copy_fields(fields)
    field_timestamp = "DATE_TIME"

    if field_timestamp not in result:
        result.append(field_timestamp)
    return ",".join(result)


def get_fields_summaries(fields, **kwargs):
    fields = fields_arg_parser.get_list(fields)
    result = copy_fields(fields)
    interval = kwargs.get("interval")
    field_timestamp = field_timestamp_by_day_interval_type.get(
        get_day_interval_type(interval or DayIntervalType.INTER)
    )
    if field_timestamp not in result:
        result.append(field_timestamp)
    return ",".join(result)


class SummariesDataProvider(HistoricalDataProvider):
    def _get_axis_name(self, interval, **kwargs):
        axis_name = axis_by_day_interval_type.get(
            get_day_interval_type(interval or DayIntervalType.INTER)
        )
        return axis_name


class EventsDataProvider(HistoricalDataProvider):
    def _get_axis_name(self, interval, **kwargs):
        return "Timestamp"



from enum import Enum, auto
from typing import TYPE_CHECKING, Union

from ._adc_context import ADCContext
from ._adc_rdp_context import ADCRDPContext
from ._adc_udf_context import ADCUDFContext
from ._cust_inst_context import CustInstContext
from ._cust_inst_rdp_context import CustInstRDPContext
from ._cust_inst_udf_context import CustInstUDFContext
from ._hp_and_cust_inst_context import HPAndCustInstContext
from ._hp_context import HPContext
from ._hp_rdp_context import HPRDPContext
from ._hp_udf_context import HPUDFContext
from ...content.fundamental_and_reference._data_grid_type import DataGridType

if TYPE_CHECKING:
    from .._containers import (
        FieldsContainer,
        UniverseContainer,
    )


class ContextType(Enum):
    ADC = auto()
    HP = auto()
    HPAndCustInst = auto()
    CustInst = auto()


data_grid_type_by_context_class_by_context_type = {
    ContextType.ADC: {
        DataGridType.UDF: ADCUDFContext,
        DataGridType.RDP: ADCRDPContext,
    },
    ContextType.HP: {
        DataGridType.UDF: HPUDFContext,
        DataGridType.RDP: HPRDPContext,
    },
    ContextType.CustInst: {
        DataGridType.UDF: CustInstUDFContext,
        DataGridType.RDP: CustInstRDPContext,
    },
    ContextType.HPAndCustInst: {
        DataGridType.UDF: HPAndCustInstContext,
        DataGridType.RDP: HPAndCustInstContext,
    },
}


def get_context(
    context_type: ContextType,
    data_grid_type: "DataGridType",
    universe: "UniverseContainer",
    fields: "FieldsContainer",
) -> Union[CustInstContext, ADCContext, HPContext]:
    data_grid_type_by_context_class = (
        data_grid_type_by_context_class_by_context_type.get(context_type)
    )

    if not data_grid_type_by_context_class:
        raise TypeError(f"Unexpected context_type. Type: {context_type}")

    context_class = data_grid_type_by_context_class.get(data_grid_type)

    if not context_class:
        raise TypeError(f"Unexpected type. Type: {data_grid_type}")

    return context_class(universe, fields)


def get_cust_inst_context(
    data_grid_type: "DataGridType",
    universe: "UniverseContainer",
    fields: "FieldsContainer",
) -> CustInstContext:
    return get_context(ContextType.CustInst, data_grid_type, universe, fields)


def get_adc_context(
    data_grid_type: "DataGridType",
    universe: "UniverseContainer",
    fields: "FieldsContainer",
) -> ADCContext:
    return get_context(ContextType.ADC, data_grid_type, universe, fields)


def get_hp_context(
    data_grid_type: "DataGridType",
    universe: "UniverseContainer",
    fields: "FieldsContainer",
) -> HPContext:
    return get_context(ContextType.HP, data_grid_type, universe, fields)


def get_hp_and_custinst_context(
    data_grid_type: "DataGridType",
    universe: "UniverseContainer",
    fields: "FieldsContainer",
) -> HPContext:
    return get_context(ContextType.HPAndCustInst, data_grid_type, universe, fields)



from typing import Optional, Union, TYPE_CHECKING

from ._data_provider import fx_swp_to_swp_data_provider as data_provider
from .._core.session import get_default
from .._tools import Copier
from ..delivery._data._data_provider_layer import _check_response

if TYPE_CHECKING:
    from ._fx_spot_quote import FxSpotQuote
    from ._fx_swap_points import FxSwapPoints
    from .._types import OptDateTime, OptStrStrs
    from ..delivery._data._response import Response


def fx_swp_to_swp(
    fx_cross_code: str,
    market_data_date_time: "OptDateTime" = None,
    tenors: "OptStrStrs" = None,
    fields: "OptStrStrs" = None,
    spot_ccy_1: Optional[Union[dict, "FxSpotQuote"]] = None,
    spot_ccy_2: Optional[Union[dict, "FxSpotQuote"]] = None,
    swap_points_ccy_1: Optional[Union[dict, "FxSwapPoints"]] = None,
    swap_points_ccy_2: Optional[Union[dict, "FxSwapPoints"]] = None,
) -> "Response":
    """
    Computes the cross currency curve using the swap points and
    spot rates of each of the currencies in the FX Cross pair against the pivot currency

    Parameters
    ----------
    fx_cross_code : str
        The currency pair of FX Cross, expressed in ISO 4217 alphabetical format (e.g., 'EURCHF').
        The user can specify a pivot currency with 3-currency ISO code in the FX Cross (e.g., 'GBPEURCHF'),
        where the second currency is the pivot currency. By default, the pivot currency is 'USD'
    market_data_date_time : OptDateTime, optional
        The date at which the market data is retrieved. The value is expressed
        in ISO 8601 format YYYY-MM-DD (e.g. '2021-01-01').
    tenors : str, list of str, optional
        An array of requested tenors, or/and end dates. the value can be expressed as
        the code indicating the time period (e.g., '1m', '6m', '4y'), or in iso 8601
        format 'yyy-mm-dd' (e.g., '2021-01-01')
    fields : str, list of str, optional
        An array of the requested fields
    spot_ccy_1 : dict or FxSpotQuote, optional

    spot_ccy_2 : dict or FxSpotQuote, optional

    swap_points_ccy_1 : dict or FxSwapPoints, optional

    swap_points_ccy_2 : dict or FxSwapPoints, optional

    Returns
    -------
    Response

    Examples
    -------
    >>> import refinitiv.data as rd
    >>> response = rd.qpl.fx_swp_to_swp(
    ...     fx_cross_code="EURUSD",
    ...     market_data_date_time="2022-09-22",
    ...     spot_ccy_1=rd.qpl.FxSpotQuote(source="D3"),
    ...     spot_ccy_2=rd.qpl.FxSpotQuote(bid=1, ask=2),
    ...     swap_points_ccy_1=rd.qpl.FxSwapPoints(
    ...         additional_tenor_types=[rd.qpl.TenorTypes.LONG],
    ...         source="ICAP",
    ...     ),
    ...     swap_points_ccy_2=rd.qpl.FxSwapPoints(
    ...         additional_tenor_types=[rd.qpl.TenorTypes.LONG, rd.qpl.TenorTypes.ODD],
    ...         source="D3",
    ...         overrides=[
    ...             rd.qpl.TenorBidAsk(tenor="1M", bid=50, ask=60),
    ...             rd.qpl.TenorBidAsk(tenor="2M", bid=90),
    ...         ],
    ...     ),
    ... )
    >>> response.data.df
    >>> response.data.raw
    """
    session = get_default()
    url = session.config.get("apis.data.qpl-functions.endpoints.fx_swp_to_swp")
    tenors = tenors and Copier.get_list(tenors)
    fields = fields and Copier.get_list(fields)
    response = data_provider.get_data(
        session,
        url,
        fx_cross_code=fx_cross_code,
        market_data_date_time=market_data_date_time,
        tenors=tenors,
        fields=fields,
        spot_ccy_1=spot_ccy_1,
        spot_ccy_2=spot_ccy_2,
        swap_points_ccy_1=swap_points_ccy_1,
        swap_points_ccy_2=swap_points_ccy_2,
    )
    _check_response(response, session.config)
    return response



from typing import Optional, TYPE_CHECKING

from ...._object_definition import ObjectDefinition
from ._enums import QuotationMode

if TYPE_CHECKING:
    from ......_types import OptStr, OptBool


class FxForwardInstrumentDefinition(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    instrument_code : str, optional
        The code used to define the instrument.
    tenor : str, optional
        The code indicating the instrument tenor (e.g., '6m', '1y').
    quotation_mode : QuotationMode, optional
        The quotation defining the price type of the instrument. the possible values
        are:   * swappoint   * outright
    is_non_deliverable : bool, optional
        True if the instrument is non deliverable
    synthetic_instrument_code : str, optional
        The code used to define the formula.
    template : str, optional
        A reference to a style used to define the instrument.
    """

    def __init__(
        self,
        instrument_code: "OptStr" = None,
        tenor: "OptStr" = None,
        quotation_mode: Optional[QuotationMode] = None,
        is_non_deliverable: "OptBool" = None,
        synthetic_instrument_code: "OptStr" = None,
        template: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.instrument_code = instrument_code
        self.tenor = tenor
        self.quotation_mode = quotation_mode
        self.is_non_deliverable = is_non_deliverable
        self.synthetic_instrument_code = synthetic_instrument_code
        self.template = template

    @property
    def quotation_mode(self):
        """
        The quotation defining the price type of the instrument. the possible values
        are:   * swappoint   * outright
        :return: enum QuotationMode
        """
        return self._get_enum_parameter(QuotationMode, "quotationMode")

    @quotation_mode.setter
    def quotation_mode(self, value):
        self._set_enum_parameter(QuotationMode, "quotationMode", value)

    @property
    def instrument_code(self):
        """
        The code used to define the instrument.
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def is_non_deliverable(self):
        """
        True if the instrument is non deliverable
        :return: bool
        """
        return self._get_parameter("isNonDeliverable")

    @is_non_deliverable.setter
    def is_non_deliverable(self, value):
        self._set_parameter("isNonDeliverable", value)

    @property
    def synthetic_instrument_code(self):
        """
        The code used to define the formula.
        :return: str
        """
        return self._get_parameter("syntheticInstrumentCode")

    @synthetic_instrument_code.setter
    def synthetic_instrument_code(self, value):
        self._set_parameter("syntheticInstrumentCode", value)

    @property
    def template(self):
        """
        A reference to a style used to define the instrument.
        :return: str
        """
        return self._get_parameter("template")

    @template.setter
    def template(self, value):
        self._set_parameter("template", value)

    @property
    def tenor(self):
        """
        The code indicating the instrument tenor (e.g., '6m', '1y').
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)



import pytest
from refinitiv.data.delivery._data._endpoint_data import EndpointData


@pytest.mark.parametrize(
    ("attrib_name", "expected_has_attrib"), [("raw", True), ("df", False)]
)
def test_endpoint_data_attrib(attrib_name, expected_has_attrib):
    # when
    testing_has_attrib = hasattr(EndpointData, attrib_name)

    assert testing_has_attrib == expected_has_attrib



from contextlib import AbstractContextManager
from typing import Callable, Iterable, List, Optional, TYPE_CHECKING, Union

import pandas as pd

from ._mixed_streams import Stream
from ._pricing_recorder import PricingRecorder
from .._core.session import get_default, raise_if_closed
from .._tools import cached_property, iterator_str_arg_parser
from ..content._universe_stream import _UniverseStream

if TYPE_CHECKING:
    from .. import OpenState


def make_callback(on_data, logger):
    def callback(update, ric, stream):
        try:
            stream = PricingStream(stream)
            df = pd.DataFrame(update, index=[ric])
            on_data(df, ric, stream)
        except Exception as error:
            logger.error(error)

    return callback


def open_pricing_stream(
    universe: Union[str, Iterable[str]],
    fields: Union[str, List[str]] = None,
    service: Optional[str] = None,
    on_data: Optional[Callable] = None,
) -> "PricingStream":
    """
    Creates and opens a pricing stream.

    Parameters
    ----------
    universe : str | Iterable[str]
        Instruments to request
    fields : str | list, optional
        Fields to request
    service : str, optional
        Name of the streaming service publishing the instruments
    on_data : function, optional
        Callback function

    Returns
    ----------
    PricingStream

    Examples
    -------
    >>> import refinitiv.data as rd
    >>> def callback(updated_data, ric, stream):
    ...    print(updated_data)
    >>> pricing_stream = rd.open_pricing_stream(universe=['EUR='], fields=['BID', 'ASK', 'OPEN_PRC'], on_data=callback)  # noqa
    """
    session = get_default()
    raise_if_closed(session)

    logger = session.logger()

    universe = iterator_str_arg_parser.get_list(universe)
    _stream = Stream(universe=universe, fields=fields, service=service)

    if on_data:
        _stream.on_update(make_callback(on_data, logger))
        _stream.on_refresh(make_callback(on_data, logger))
    stream = PricingStream(_stream)

    stream.open()
    return stream


class PricingStream(AbstractContextManager):
    def __init__(self, stream):
        self._stream = stream

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def open(self, with_updates: bool = True) -> "OpenState":
        return self._stream.open(with_updates=with_updates)

    def close(self) -> "OpenState":
        return self._stream.close()

    def get_snapshot(
        self,
        universe: Union[str, List[str], None] = None,
        fields: Optional[List[str]] = None,
        convert: bool = True,
    ) -> "pd.DataFrame":
        return self._stream.get_snapshot(
            universe=universe, fields=fields, convert=convert
        )

    def _get_fields(self, universe: str, fields: Optional[list] = None) -> dict:
        return self._stream._get_fields(universe=universe, fields=fields)

    def add_instruments(self, *args):
        self._stream.add_instruments(*args)

    def remove_instruments(self, *args):
        self._stream.remove_instruments(*args)

    def __getitem__(self, item) -> "_UniverseStream":
        return self._stream.__getitem__(item)

    def __iter__(self):
        return self._stream.__iter__()

    @cached_property
    def recorder(self) -> PricingRecorder:
        return PricingRecorder(self._stream._stream)



from tests.unit.conftest import StubResponse

# fmt: off
INTRA_WITH_DATES_0_1 = [
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    [
                        "2022-05-25T23:59:00.000000000Z",
                        90.75,
                        90.7,
                        90.75,
                        90.7,
                        17,
                        2289,
                    ],
                    [
                        "2022-05-25T23:58:00.000000000Z",
                        90.75,
                        90.7,
                        90.74,
                        90.72,
                        20,
                        1778,
                    ],
                    [
                        "2022-05-25T23:57:00.000000000Z",
                        90.75,
                        90.7,
                        90.72,
                        90.74,
                        22,
                        3790,
                    ],
                ],
            }
        ]
    ),
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    [
                        "2022-05-25T23:56:00.000000000Z",
                        90.8,
                        90.66,
                        90.7,
                        90.73,
                        19,
                        2721,
                    ],
                    [
                        "2022-05-25T23:55:00.000000000Z",
                        90.74,
                        90.68,
                        90.74,
                        90.7,
                        16,
                        1525,
                    ],
                    [
                        "2022-05-25T23:54:00.000000000Z",
                        90.77,
                        90.68,
                        90.7,
                        90.74,
                        22,
                        4193,
                    ],
                ],
            }
        ]
    ),
]


INTRA_WITH_DATES_2 = [
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    [
                        "2022-05-25T23:59:00.000000000Z",
                        90.75,
                        90.7,
                        90.75,
                        90.7,
                        17,
                        2289,
                    ],
                    [
                        "2022-05-25T23:58:00.000000000Z",
                        90.75,
                        90.7,
                        90.74,
                        90.72,
                        20,
                        1778,
                    ],
                    [
                        "2022-05-25T23:57:00.000000000Z",
                        90.75,
                        90.7,
                        90.72,
                        90.74,
                        22,
                        3790,
                    ],
                    [
                        "2022-05-25T23:56:00.000000000Z",
                        90.8,
                        90.66,
                        90.7,
                        90.73,
                        19,
                        2721,
                    ],
                    [
                        "2022-05-25T23:55:00.000000000Z",
                        90.74,
                        90.68,
                        90.74,
                        90.7,
                        16,
                        1525,
                    ],
                    [
                        "2022-05-25T23:54:00.000000000Z",
                        90.77,
                        90.68,
                        90.7,
                        90.74,
                        22,
                        4193,
                    ],
                ],
            }
        ]
    )
]


INTRA_WITH_DATES_3 = [
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    [
                        "2022-05-25T23:59:00.000000000Z",
                        90.75,
                        90.7,
                        90.75,
                        90.7,
                        17,
                        2289,
                    ],
                    [
                        "2022-05-25T23:58:00.000000000Z",
                        90.75,
                        90.7,
                        90.74,
                        90.72,
                        20,
                        1778,
                    ],
                    [
                        "2022-05-25T23:57:00.000000000Z",
                        90.75,
                        90.7,
                        90.72,
                        90.74,
                        22,
                        3790,
                    ],
                ],
            }
        ]
    )
]


INTRA_WITH_DATES_4 = [
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    [
                        "2022-05-25T23:57:00.000000000Z",
                        90.75,
                        90.7,
                        90.72,
                        90.74,
                        22,
                        3790,
                    ],
                ],
            }
        ]
    ),
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    [
                        "2022-05-25T23:56:00.000000000Z",
                        90.8,
                        90.66,
                        90.7,
                        90.73,
                        19,
                        2721,
                    ],
                    [
                        "2022-05-25T23:55:00.000000000Z",
                        90.74,
                        90.68,
                        90.74,
                        90.7,
                        16,
                        1525,
                    ],
                    [
                        "2022-05-25T23:54:00.000000000Z",
                        90.77,
                        90.68,
                        90.7,
                        90.74,
                        22,
                        4193,
                    ],
                ],
            }
        ]
    ),
]

INTRA_WITH_COUNT_0 = [
    StubResponse(
        [
            {
                "universe": {"ric": "LSEG.L"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "BID", "type": "number", "decimalChar": "."},
                    {"name": "ASK", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    ["2022-10-18T04:00:00.000000000Z", 6872, 9000],
                    ["2022-10-17T16:30:00.000000000Z", 6872, 9000],
                    ["2022-10-17T15:40:00.000000000Z", 7312, 7328],
                    ["2022-10-17T15:35:00.000000000Z", 7312, 7324],
                    ["2022-10-17T15:34:00.000000000Z", 8312, 6664],
                ],
                "meta": {
                    "blendingEntry": {
                        "headers": [
                            {"name": "COLLECT_DATETIME", "type": "string"},
                            {"name": "RTL", "type": "number", "decimalChar": "."},
                            {"name": "SOURCE_DATETIME", "type": "string"},
                            {"name": "SEQNUM", "type": "string"},
                        ],
                        "data": [
                            [
                                "2022-10-18T04:00:09.113000000Z",
                                21920,
                                "2022-10-18T04:00:09.113000000Z",
                                "185",
                            ]
                        ],
                    }
                },
            }
        ]
    )
]

INTRA_EMPTY_DATA = [
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [],
            }
        ]
    )
]

INTRA_TWO_UNIVERSES_EMPTY_DATA = [
    StubResponse(
        [
            {
                "universe": {"ric": "AMD.O"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [],
            }
        ]
    ),
    StubResponse(
        [
            {
                "universe": {"ric": "VOD.L"},
                "interval": "PT1M",
                "summaryTimestampLabel": "startPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "HIGH_1", "type": "number", "decimalChar": "."},
                    {"name": "LOW_1", "type": "number", "decimalChar": "."},
                    {"name": "OPEN_PRC", "type": "number", "decimalChar": "."},
                    {"name": "TRDPRC_1", "type": "number", "decimalChar": "."},
                    {"name": "NUM_MOVES", "type": "number", "decimalChar": "."},
                    {"name": "ACVOL_UNS", "type": "number", "decimalChar": "."},
                ],
                "data": [],
            }
        ]
    ),
]

INST_IS_BAD = [
    StubResponse(
        [
            {
                "universe": {"ric": "LSEG.L.ffff"},
                "status": {
                    "code": "TS.Intraday.UserRequestError.90001",
                    "message": "The universe is not found.",
                },
            }
        ]
    )
]

INSTS_ARE_BAD = [
    StubResponse(
        [
            {
                "universe": {"ric": "LSEG.L.ffff"},
                "status": {
                    "code": "TS.Intraday.UserRequestError.90001",
                    "message": "The universe is not found.",
                },
            }
        ]
    ),
    StubResponse(
        [
            {
                "universe": {"ric": "VOD.Lffff"},
                "status": {
                    "code": "TS.Intraday.UserRequestError.90001",
                    "message": "The universe is not found.",
                },
            }
        ]
    ),
]


EVENT_WITH_DATES_0_1 = [
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}], "data": [
                ["2022-10-18T09:17:04.466000000Z", "quote", 26142, 1.1319, 1.132,
                 1.13195, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
                ["2022-10-18T09:17:04.416000000Z", "quote", 26078, 1.1315, 1.132,
                 1.13175, None, "CKLU", "CKLU", "StoneX Finan", "LUX", None],
                ["2022-10-18T09:17:04.310000000Z", "quote", 26014, 1.1318, 1.132,
                 1.1319, None, None, "ZKBZ", "ZUERCHER KB", "ZUR", None],
                ["2022-10-18T09:17:04.118000000Z", "quote", 25950, 1.1318, 1.1322,
                 1.132, None, "BCFX", None, "BARCLAYS", "LON", None],
                ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321,
                 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None]], "meta": {
                "blendingEntry": {
                    "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                                {"name": "RTL", "type": "number", "decimalChar": "."}],
                    "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    )
]


EVENT_WITH_DATES_2 = [
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
                ["2022-10-18T09:17:04.466000000Z", "quote", 26142, 1.1319, 1.132, 1.13195, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
                ["2022-10-18T09:17:04.416000000Z", "quote", 26078, 1.1315, 1.132, 1.13175, None, "CKLU", "CKLU", "StoneX Finan", "LUX", None],
                ["2022-10-18T09:17:04.310000000Z", "quote", 26014, 1.1318, 1.132, 1.1319, None, None, "ZKBZ", "ZUERCHER KB", "ZUR", None],
                ["2022-10-18T09:17:04.118000000Z", "quote", 25950, 1.1318, 1.1322, 1.132, None, "BCFX", None, "BARCLAYS", "LON", None],
                ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321, 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None]
          ],
          "meta": {
                "blendingEntry": {
                    "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                                {"name": "RTL", "type": "number", "decimalChar": "."}],
                    "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    ),
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
              ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321, 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None],
              ["2022-10-18T09:17:04.270000000Z", "quote", 26142, 1.1315, 1.132, 1.13175, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
          ],
          "meta": {
              "blendingEntry": {
                  "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                              {"name": "RTL", "type": "number", "decimalChar": "."}],
                  "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    )
]

EVENT_WITH_COUNT_0 = [
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
                ["2022-10-18T09:59:59.575000000Z", "quote", 46110, 1.1271, 1.1275, 1.1273, None, "BCFX", None, "BARCLAYS", "LON", None],
                ["2022-10-18T09:59:58.574000000Z", "quote", 46046, 1.1271, 1.1275, 1.1273, None, "BCFX", None, "BARCLAYS", "LON", None],
                ["2022-10-18T09:59:57.487000000Z", "quote", 45982, 1.1271, 1.1275, 1.1273, None, "SAHK", "SAHK", "SANTANDER", "HKG", None]
          ]}]
    )
]


EVENT_WITH_COUNT_1 = [
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
                ["2022-10-18T09:17:04.466000000Z", "quote", 26142, 1.1319, 1.132, 1.13195, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
                ["2022-10-18T09:17:04.416000000Z", "quote", 26078, 1.1315, 1.132, 1.13175, None, "CKLU", "CKLU", "StoneX Finan", "LUX", None],
                ["2022-10-18T09:17:04.310000000Z", "quote", 26014, 1.1318, 1.132, 1.1319, None, None, "ZKBZ", "ZUERCHER KB", "ZUR", None],
                ["2022-10-18T09:17:04.118000000Z", "quote", 25950, 1.1318, 1.1322, 1.132, None, "BCFX", None, "BARCLAYS", "LON", None],
                ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321, 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None]
          ],
          "meta": {
                "blendingEntry": {
                    "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                                {"name": "RTL", "type": "number", "decimalChar": "."}],
                    "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    ),
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
              ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321, 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None],
              ["2022-10-18T09:17:04.270000000Z", "quote", 26142, 1.1315, 1.132, 1.13175, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
          ],
          "meta": {
              "blendingEntry": {
                  "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                              {"name": "RTL", "type": "number", "decimalChar": "."}],
                  "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    )
]


EVENT_WITH_COUNT_2 = [
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
                ["2022-10-18T09:17:04.466000000Z", "quote", 26142, 1.1319, 1.132, 1.13195, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
                ["2022-10-18T09:17:04.416000000Z", "quote", 26078, 1.1315, 1.132, 1.13175, None, "CKLU", "CKLU", "StoneX Finan", "LUX", None],
                ["2022-10-18T09:17:04.310000000Z", "quote", 26014, 1.1318, 1.132, 1.1319, None, None, "ZKBZ", "ZUERCHER KB", "ZUR", None],
                ["2022-10-18T09:17:04.118000000Z", "quote", 25950, 1.1318, 1.1322, 1.132, None, "BCFX", None, "BARCLAYS", "LON", None],
                ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321, 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None]
          ],
          "meta": {
                "blendingEntry": {
                    "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                                {"name": "RTL", "type": "number", "decimalChar": "."}],
                    "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    ),
    StubResponse(
        [{"universe": {"ric": "GBP="},
          "adjustments": ["exchangeCorrection", "manualCorrection"],
          "defaultPricingField": "BID",
          "headers": [{"name": "DATE_TIME", "type": "string"},
                      {"name": "EVENT_TYPE", "type": "string"},
                      {"name": "RTL", "type": "number", "decimalChar": "."},
                      {"name": "BID", "type": "number", "decimalChar": "."},
                      {"name": "ASK", "type": "number", "decimalChar": "."},
                      {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                      {"name": "DSPLY_NAME", "type": "string"},
                      {"name": "SRC_REF1", "type": "string"},
                      {"name": "DLG_CODE1", "type": "string"},
                      {"name": "CTBTR_1", "type": "string"},
                      {"name": "CTB_LOC1", "type": "string"},
                      {"name": "QUALIFIERS", "type": "string"}],
          "data": [
              ["2022-10-18T09:17:03.269000000Z", "quote", 25886, 1.1318, 1.1321, 1.13195, None, "COB1", "CBKF", "Commerzbank", "FFT", None],
              ["2022-10-18T09:17:04.270000000Z", "quote", 26142, 1.1315, 1.132, 1.13175, None, "SAHK", "SAHK", "SANTANDER", "HKG", None],
          ],
          "meta": {
              "blendingEntry": {
                  "headers": [{"name": "COLLECT_DATETIME", "type": "string"},
                              {"name": "RTL", "type": "number", "decimalChar": "."}],
                  "data": [["2022-10-18T09:17:04.466000000Z", 26142]]}}}]
    )
]



import pytest
from refinitiv.data.content.estimates._enums import Package

success_param = [
    ({"universe": "test_value", "package": Package.BASIC}),
    ({"universe": "test_value", "package": Package.STANDARD}),
    ({"universe": "test_value", "package": Package.PROFESSIONAL}),
    (
        {
            "universe": "test_value",
            "package": Package.BASIC,
            "use_field_names_in_headers": False,
        }
    ),
    (
        {
            "universe": "test_value",
            "package": Package.STANDARD,
            "use_field_names_in_headers": False,
        }
    ),
    (
        {
            "universe": "test_value",
            "package": Package.PROFESSIONAL,
            "use_field_names_in_headers": False,
        }
    ),
    (
        {
            "universe": "test_value",
            "package": Package.BASIC,
            "use_field_names_in_headers": True,
        }
    ),
    (
        {
            "universe": "test_value",
            "package": Package.STANDARD,
            "use_field_names_in_headers": True,
        }
    ),
    (
        {
            "universe": "test_value",
            "package": Package.PROFESSIONAL,
            "use_field_names_in_headers": True,
        }
    ),
]

error_param = [
    ({"universe": "test_value", "package": 1}),
    ({"universe": "test_value", "package": []}),
    ({"universe": "test_value", "package": {}}),
    (
        {
            "universe": "test_value",
            "package": Package.BASIC,
            "use_field_names_in_headers": 1,
        }
    ),
    (
        {
            "universe": "test_value",
            "package": Package.BASIC,
            "use_field_names_in_headers": "true",
        }
    ),
]

success_param_kpi = [
    ({"universe": "test_value"}),
    ({"universe": "test_value", "use_field_names_in_headers": False}),
    ({"universe": "test_value", "use_field_names_in_headers": True}),
]

error_param_kpi = [
    ({"universe": "test_value", "package": 1}),
    ({"universe": "test_value", "package": []}),
    ({"universe": "test_value", "package": {}}),
    ({"universe": "test_value", "use_field_names_in_headers": 1}),
    ({"universe": "test_value", "use_field_names_in_headers": "true"}),
]


@pytest.fixture(params=success_param)
def create_estimates_definition(request):
    try:
        yield request.param
    except Exception as e:
        assert False, str(e)


@pytest.fixture(params=error_param)
def create_estimates_definition_with_error(request):
    return request.param


@pytest.fixture(params=success_param_kpi)
def create_estimates_kpi_definition(request):
    try:
        yield request.param
    except Exception as e:
        assert False, str(e)


@pytest.fixture(params=error_param_kpi)
def create_estimates_kpi_definition_with_error(request):
    return request.param



# coding: utf8

__all__ = ["PdeParameters"]

from .._object_definition import ObjectDefinition


class PdeParameters(ObjectDefinition):
    def __init__(
        self,
        pde_space_step_number=None,
        pde_standard_deviation=None,
        pde_time_step_number=None,
    ):
        super().__init__()
        self.pde_space_step_number = pde_space_step_number
        self.pde_standard_deviation = pde_standard_deviation
        self.pde_time_step_number = pde_time_step_number

    @property
    def pde_space_step_number(self):
        """
        :return: int
        """
        return self._get_parameter("pdeSpaceStepNumber")

    @pde_space_step_number.setter
    def pde_space_step_number(self, value):
        self._set_parameter("pdeSpaceStepNumber", value)

    @property
    def pde_standard_deviation(self):
        """
        :return: int
        """
        return self._get_parameter("pdeStandardDeviation")

    @pde_standard_deviation.setter
    def pde_standard_deviation(self, value):
        self._set_parameter("pdeStandardDeviation", value)

    @property
    def pde_time_step_number(self):
        """
        :return: int
        """
        return self._get_parameter("pdeTimeStepNumber")

    @pde_time_step_number.setter
    def pde_time_step_number(self, value):
        self._set_parameter("pdeTimeStepNumber", value)



from typing import List, Union, Dict

from ._context import UDFMixin
from ._cust_inst_context import CustInstContext


class CustInstUDFContext(UDFMixin, CustInstContext):
    def get_headers(self, headers: dict) -> List[str]:
        return [
            item["name"].capitalize() if item["name"] == "DATE" else item["name"]
            for item in headers
        ]

    def _parse_list_to_data(self, raw: list, field_to_idx: dict) -> List[List[dict]]:
        data = []

        for raw_item in raw:
            if not raw_item:
                continue

            ric = raw_item["universe"]["ric"]
            headers = self.get_headers(raw_item["headers"])

            for raw_item_data in raw_item["data"]:
                template = {"Instrument": ric}
                template.update(
                    {header: data for header, data in zip(headers, raw_item_data)}
                )
                item = []
                for field in field_to_idx:
                    item.insert(field_to_idx[field], template.get(field))

                data.append(item)

        return data

    def _parse_dict_to_data(self, raw: dict, field_to_idx: dict) -> List[List[dict]]:
        data = []

        ric = raw["universe"]["ric"]
        headers = self.get_headers(raw["headers"])

        for raw_data_item in raw["data"]:
            template = {"Instrument": ric}
            template.update(
                {header: data for header, data in zip(headers, raw_data_item)}
            )
            item = []
            for field in field_to_idx:
                item.insert(field_to_idx[field], template.get(field))

            data.append(item)

        return data

    def prepare_data(self, raw, fields, *args) -> List[List[dict]]:
        field_to_idx = {"Instrument": 0, "Date": 1}
        field_to_idx.update({item: fields.index(item) + 2 for item in fields})

        if isinstance(raw, dict):
            return self._parse_dict_to_data(raw, field_to_idx)

        elif isinstance(raw, list):
            return self._parse_list_to_data(raw, field_to_idx)

    def prepare_headers(self, raw: Union[list, dict], *args) -> List[List[Dict]]:
        headers = [{"displayName": "Instrument"}, {"displayName": "Date"}]

        for item in raw:
            headers.append({"displayName": item, "field": item})

        return [headers]



# coding: utf8


from typing import Optional

from .._instrument_definition import ObjectDefinition


class RepoParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    coupon_paid_at_horizon : bool, optional
        Flag that defines whether coupons paid at horizon.  this has no impact on
        pricing.
    haircut_rate_percent : float, optional
        The reduction applied to the value of an underlying asset for purposes of
        calculating a repo collateral. the value is computed as [(initialmarginpercent -
        100) / initialmarginpercent] and expressed in percentages. either haircut or
        initial marging field can be bet. optional. by default it is computed from
        initialmarginpercent.
    initial_margin_percent : float, optional
        The initial market value of collateral expressed as a percentage of the purchase
        price of the underlying asset. either haircutratepercent or initialmarginpercent
        can be overriden. optional. default value is 100.
    purchase_price : float, optional
        Purchase price of the asset. this parameter can be used to solve repurchaseprice
        from this purchaseprice value. optional. by default it is computed from net
        present value and initial margin.
    repurchase_price : float, optional
        Repurchase price of the asset. this parameter can be used to solve purchaseprice
        from this repurchaseprice value. optional. by default it is computed from
        underlying end price, or solved from purchaseprice and repo rate.
    """

    def __init__(
        self,
        coupon_paid_at_horizon: Optional[bool] = None,
        haircut_rate_percent: Optional[float] = None,
        initial_margin_percent: Optional[float] = None,
        purchase_price: Optional[float] = None,
        repurchase_price: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.coupon_paid_at_horizon = coupon_paid_at_horizon
        self.haircut_rate_percent = haircut_rate_percent
        self.initial_margin_percent = initial_margin_percent
        self.purchase_price = purchase_price
        self.repurchase_price = repurchase_price

    @property
    def coupon_paid_at_horizon(self):
        """
        Flag that defines whether coupons paid at horizon.  this has no impact on
        pricing.
        :return: bool
        """
        return self._get_parameter("couponPaidAtHorizon")

    @coupon_paid_at_horizon.setter
    def coupon_paid_at_horizon(self, value):
        self._set_parameter("couponPaidAtHorizon", value)

    @property
    def haircut_rate_percent(self):
        """
        The reduction applied to the value of an underlying asset for purposes of
        calculating a repo collateral. the value is computed as [(initialmarginpercent -
        100) / initialmarginpercent] and expressed in percentages. either haircut or
        initial marging field can be bet. optional. by default it is computed from
        initialmarginpercent.
        :return: float
        """
        return self._get_parameter("haircutRatePercent")

    @haircut_rate_percent.setter
    def haircut_rate_percent(self, value):
        self._set_parameter("haircutRatePercent", value)

    @property
    def initial_margin_percent(self):
        """
        The initial market value of collateral expressed as a percentage of the purchase
        price of the underlying asset. either haircutratepercent or initialmarginpercent
        can be overriden. optional. default value is 100.
        :return: float
        """
        return self._get_parameter("initialMarginPercent")

    @initial_margin_percent.setter
    def initial_margin_percent(self, value):
        self._set_parameter("initialMarginPercent", value)

    @property
    def purchase_price(self):
        """
        Purchase price of the asset. this parameter can be used to solve repurchaseprice
        from this purchaseprice value. optional. by default it is computed from net
        present value and initial margin.
        :return: float
        """
        return self._get_parameter("purchasePrice")

    @purchase_price.setter
    def purchase_price(self, value):
        self._set_parameter("purchasePrice", value)

    @property
    def repurchase_price(self):
        """
        Repurchase price of the asset. this parameter can be used to solve purchaseprice
        from this repurchaseprice value. optional. by default it is computed from
        underlying end price, or solved from purchaseprice and repo rate.
        :return: float
        """
        return self._get_parameter("repurchasePrice")

    @repurchase_price.setter
    def repurchase_price(self, value):
        self._set_parameter("repurchasePrice", value)



import re

from ...._tools._datetime import Converter

datetime_converter = Converter()
regexp_filename_obj = re.compile(
    r".*([Dd]elta|[Ii]nit).*(\d{4}-\d{2}-\d{2}).*(\.jsonl.gz|\.jsonl)$"
)
regexp_init_archives = re.compile(r".*(-[Ii]nit-).*\d{4}-\d{2}-\d{2}.*(.gz)$")
regexp_delta_files = re.compile(r".*(Jsonl-[Dd]elta-).*\d{4}-\d{2}-\d{2}.*")
regexp_init_file = re.compile(r".*(Jsonl-[Ii]nit-).*\d{4}-\d{2}-\d{2}.*")
regexp_jsonl_fileset = re.compile(r".*(-Jsonl-).*\d{4}-\d{2}-\d{2}.*")
regexp_filename_date = re.compile(r".*(\d{4}-\d{2}-\d{2}).*")


def is_correct_filename(filename):
    return regexp_filename_obj.match(filename) is not None


def is_init_file(filename):
    return regexp_init_file.match(filename) is not None


def get_date_from_filename(filename):
    result = re.search(r"\d{4}-\d{2}-\d{2}", filename)
    return result.group(0)


def get_datetime_from_filename(filename):
    date = get_date_from_filename(filename)
    result = datetime_converter.convert(date)
    return result


def get_the_newest_file(filenames):
    sorted_filenames = sorted(filenames, key=lambda a: get_datetime_from_filename(a))
    if not sorted_filenames:
        return None
    result = sorted_filenames[-1]
    return result


def get_init_archives(filenames):
    # return init archives with correct name
    return filter(lambda a: regexp_init_archives.match(a), filenames)


def get_filesets_with_delta_files(file_sets):
    # return init archives with correct name
    result = [
        file_set for file_set in file_sets if regexp_delta_files.match(file_set["name"])
    ]
    return result


def get_filesets_with_the_newest_init_file_and_delta_files(file_sets):
    file_sets = [
        file_set
        for file_set in file_sets
        if regexp_jsonl_fileset.match(file_set["name"])
    ]
    init_filesets = filter(lambda a: "init" in a["name"].lower(), file_sets)
    init_dates = [get_date_from_filename(i["name"]) for i in init_filesets]
    init_dates.sort()
    if not init_dates:
        return []
    init_date = datetime_converter.convert(init_dates[-1])
    result = filter(
        lambda a: init_date
        <= datetime_converter.convert(get_date_from_filename(a["name"])),
        file_sets,
    )
    return result


def sorted_files(files):
    if not files:
        return []

    result = sorted(files)
    last_item = result[-1]

    if "init" in last_item or "Init" in last_item:
        result.insert(0, result.pop(-1))

    return result


def join_path_to_field(*args):
    if len(args) == 1:
        return args[0]

    return ".".join(filter(lambda a: bool(a), args))



from types import SimpleNamespace
from unittest.mock import mock_open
from unittest.mock import patch

import pytest


def create_stub_response(*args, **kwargs):
    response = SimpleNamespace()
    response.ok = True
    response.close = lambda: None
    response.iter_content = lambda: []
    return response


@pytest.fixture(scope="module", autouse=True)
def mock_download():
    os_makedirs_mock = patch("os.makedirs", new=lambda *_: None)
    requests_get_mock = patch("requests.get", new=create_stub_response)
    gzip_open_mock = patch("gzip.open", new=mock_open())
    file_downloader_open_mock = patch(
        "refinitiv.data.delivery.cfs._file_downloader.open", new=mock_open()
    )
    cfs_unpacker_open_mock = patch(
        "refinitiv.data.delivery.cfs._unpacker.open", new=mock_open()
    )

    os_makedirs_mock.start()
    cfs_unpacker_open_mock.start()
    requests_get_mock.start()
    gzip_open_mock.start()
    file_downloader_open_mock.start()

    yield

    os_makedirs_mock.stop()
    cfs_unpacker_open_mock.stop()
    requests_get_mock.stop()
    gzip_open_mock.stop()
    file_downloader_open_mock.stop()



import pytest

from refinitiv.data.content.ipa._enums import Axis
from refinitiv.data.content.ipa._surfaces._surfaces_data_provider import (
    SurfaceResponseFactory,
    SurfacesContentValidator,
    get_type_by_axis,
    parse_axis,
    OneSurfaceData,
)
from refinitiv.data.delivery._data._data_provider import ParsedData
from .conftest import StubDefinition


def test_surface_response_factory_create_success():
    # given
    input_value = ParsedData({}, {}, [{}])
    universe = [StubDefinition(axis=None)]
    response_factory = SurfaceResponseFactory()

    # when
    testing_value = response_factory.create_success(input_value, universe=universe)

    # then
    assert testing_value is not None


def test_data_class_surface_property():
    # given
    input_value = {"data": [{"surface": [[1], [2]]}]}
    data = OneSurfaceData(input_value, [(Axis.TENOR, Axis.TENOR)])

    # when
    surface = data.surface

    # then
    assert surface is not None


def test_curves_and_surfaces_content_validator_validate_content_data_is_false():
    # given
    input_value = ParsedData({}, {}, {"data": [{"error": {"status": "Error"}}]})
    expected_value = False
    validator = SurfacesContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


def test_get_type_by_axis_raise_error_if__cannot_find():
    with pytest.raises(ValueError, match="Cannot find axis's values type for axis"):
        get_type_by_axis(None)


def test_parse_axis_raise_error_if_axis_dont_passed():
    universe = {"surface": [[1], [2]]}
    with pytest.raises(ValueError, match="Cannot parse surface without information"):
        parse_axis(universe, None, None)



from datetime import datetime, timedelta

import allure
import numpy
import pandas as pd
import pytest
from pandas._libs.tslibs.offsets import CustomBusinessDay

from refinitiv.data import dates_and_calendars
from tests.integration.helpers import (
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)
from refinitiv.data._fin_coder_layer.dates_and_calendars._count_periods import (
    CountedPeriods,
)

end_date = datetime.utcnow()
start_date = end_date - timedelta(days=130)


@allure.suite("FinCoder layer")
@allure.feature("FinCoder - Date and Calendars")
@allure.severity(allure.severity_level.CRITICAL)
class TestDateAndCalendars:
    @allure.title("Create add_periods and get data")
    @pytest.mark.caseid("C53990671")
    def test_add_periods_and_get_data(self, open_session):
        response = dates_and_calendars.add_periods(
            start_date=start_date,
            period="4D",
            calendars=["BAR", "KOR", "JAP"],
            currencies=["USD"],
            date_moving_convention="NextBusinessDay",
            end_of_month_convention="Last28",
        )

        assert isinstance(response, numpy.datetime64)

    @allure.title("Create count_periods and get data")
    @pytest.mark.caseid("C53990673")
    def test_count_periods_and_get_data(self, open_session):
        response = dates_and_calendars.count_periods(
            start_date=timedelta(-30),
            end_date=end_date,
            period_type=dates_and_calendars.PeriodType.DAY,
            calendars=["EMU"],
            currencies=["USD"],
            day_count_basis=dates_and_calendars.DayCountBasis.DCB_ACTUAL_ACTUAL,
        )

        assert isinstance(response, CountedPeriods)
        assert isinstance(response.count, float)
        assert response.tenor is not None

    @allure.title("Create date_schedule and get data")
    @pytest.mark.caseid("C53990674")
    def test_date_schedule_and_get_data(self, open_session):
        response = dates_and_calendars.date_schedule(
            start_date=start_date,
            calendars=["EMU"],
            count=10,
            frequency=dates_and_calendars.DateScheduleFrequency.WEEKLY,
            day_of_week=dates_and_calendars.DayOfWeek.MONDAY,
        )

        for date in response:
            assert isinstance(date, numpy.datetime64)

    @allure.title("Create is_working_day and get data")
    @pytest.mark.caseid("C53990675")
    def test_is_working_day_and_get_data(self, open_session):
        response = dates_and_calendars.is_working_day(
            date="22-09-01",
            currencies=["UAH"],
            calendars=["UKR", "FRA"],
        )

        assert type(response) == bool
        assert not response

    @allure.title("Create holidays and check data with offset")
    @pytest.mark.parametrize(
        "expected_columns,expected_calendars,expected_bd_range",
        [
            (
                [
                    "name",
                    "calendars",
                    "countries",
                    "tag",
                    "date",
                ],
                ["UKR", "FRA", "USA", "EMU"],
                [
                    "2019-12-02",
                    "2019-12-03",
                    "2019-12-04",
                    "2019-12-05",
                    "2019-12-06",
                    "2019-12-09",
                    "2019-12-10",
                ],
            )
        ],
    )
    @pytest.mark.caseid("C53990690")
    def test_holidays_and_get_data(
        self, open_session, expected_columns, expected_calendars, expected_bd_range
    ):
        response = dates_and_calendars.holidays(
            start_date="2018-12-31",
            end_date="2020-01-01",
            calendars=["UKR", "FRA"],
            currencies=["USD", "EUR"],
        )

        usa_bd_range = pd.date_range(
            start="2019-12-01", end="2019-12-10", freq=response.calendars["USA"].offset
        )

        for date in usa_bd_range:
            date = date.strftime("%Y-%m-%d")
            assert (
                date in expected_bd_range
            ), f"Date {date} is not in the expected date range"

        for calendars in expected_calendars:
            assert (
                calendars in response.calendars
            ), f"{calendars} doesn`t exist in the response"

        assert isinstance(response.offset, CustomBusinessDay)
        check_response_dataframe_contains_columns_names(response, expected_columns)
        check_dataframe_column_date_for_datetime_type(response)



from typing import TYPE_CHECKING

from .._content_data import Data
from .._content_provider_layer import ContentUsageLoggerMixin
from ..._content_type import ContentType
from ..._tools import create_repr, validate_bool_value, try_copy_to_list
from ...delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ..._types import StrStrings


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]],
    DataProviderLayer[BaseResponse[Data]],
):
    """
    This class describe parameters to retrieve ESG basic data.

    Parameters
    ----------
    universe : str, list of str
        The Universe parameter allows the user to define the company they
        want content returned for, ESG content is delivered at the Company Level.

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    Examples
    --------
    >>> from refinitiv.data.content import esg
    >>> definition = esg.basic_overview.Definition("IBM.N")
    >>> response = definition.get_data()

    >>> response = await definition.get_data_async()
    """

    _USAGE_CLS_NAME = "ESG.BasicOverviewDefinition"

    def __init__(
        self,
        universe: "StrStrings",
        use_field_names_in_headers: bool = False,
    ):
        validate_bool_value(use_field_names_in_headers)
        universe = try_copy_to_list(universe)

        super().__init__(
            ContentType.ESG_BASIC_OVERVIEW,
            universe=universe,
            use_field_names_in_headers=use_field_names_in_headers,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="basic_overview",
            content=f"{{universe='{self._kwargs.get('universe')}'}}",
        )



from typing import TYPE_CHECKING

from .._base_definition_mixin import BaseDefinitionMixin


if TYPE_CHECKING:
    from ..._types import OptMainConstituentAssetClass, OptRiskType
    from ......._types import OptStr, OptBool


class CrossCurrencyCurveDefinitionDescription(BaseDefinitionMixin):
    """
    Create a cross currency curve definition

    Parameters
    ----------
    main_constituent_asset_class : MainConstituentAssetClass, optional
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
    risk_type : RiskType, optional
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
    base_currency : str, optional
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
    definition_expiry_date : str, optional
        The date after which curvedefinitions can not be used. the value is expressed in
        iso 8601 format: yyyy-mm-dd (e.g., '2021-01-01').
    is_fallback_for_fx_curve_definition : bool, optional
        The indicator used to define the fallback logic for the fx curve definition. the
        possible values are:   * true: there's a fallback logic to use cross currency
        curve definition for fx curve definition,   * false: there's no fallback logic
        to use cross currency curve definition for fx curve definition.
    is_non_deliverable : bool, optional
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
    name : str, optional
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
    quoted_currency : str, optional
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
    source : str, optional
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
    """

    def __init__(
        self,
        main_constituent_asset_class: "OptMainConstituentAssetClass" = None,
        risk_type: "OptRiskType" = None,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        definition_expiry_date: "OptStr" = None,
        is_fallback_for_fx_curve_definition: "OptBool" = None,
        is_non_deliverable: "OptBool" = None,
        name: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        source: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.main_constituent_asset_class = main_constituent_asset_class
        self.risk_type = risk_type
        self.base_currency = base_currency
        self.base_index_name = base_index_name
        self.definition_expiry_date = definition_expiry_date
        self.is_fallback_for_fx_curve_definition = is_fallback_for_fx_curve_definition
        self.is_non_deliverable = is_non_deliverable
        self.name = name
        self.quoted_currency = quoted_currency
        self.quoted_index_name = quoted_index_name
        self.source = source



import allure
import pytest

from refinitiv.data.content.esg import full_measures
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_TITLES = [
    "Instrument",
    "Period End Date",
    "TRESG Combined Score",
    "TRESG Score",
    "ESG Controversies Score",
    "Environment Pillar Score",
    "Social Pillar Score",
    "Governance Pillar Score",
    "Resource Use Score",
    "Emissions Score",
    "Accounting Controversies",
    "CSR Sustainability Committee",
    "Integrated Strategy in MD&A",
    "Global Compact Signatory",
    "Stakeholder Engagement",
    "CSR Sustainability Reporting",
    "GRI Report Guidelines",
    "CSR Sustainability Report Global Activities",
    "CSR Sustainability External Audit",
    "ESG Reporting Scope",
]


@allure.suite("Content object - ESG Full Measures")
@allure.feature("Content object - ESG Full Measures")
@allure.severity(allure.severity_level.CRITICAL)
class TestESGFullMeasures:
    @allure.title("Create ESG Full Measures definition object with valid params")
    @pytest.mark.parametrize(
        "universe,start,end",
        [("BNPP.PA", 0, -10), (["IBM.N", "VOD.L"], None, -2)],
    )
    @pytest.mark.caseid("C40232227")
    @pytest.mark.smoke
    def test_esg_full_measures_definition_object_with_valid_params_and_get_data(
        self, universe, start, end, open_platform_session
    ):
        response = full_measures.Definition(
            universe=universe, start=start, end=end
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, EXPECTED_TITLES)
        check_response_dataframe_instrument_contains_expected_universes(
            response, universe
        )
        check_non_empty_response_data(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create ESG Full Measures definition object with invalid params and get error"
    )
    @pytest.mark.caseid("C40232228")
    @pytest.mark.parametrize(
        "universe,expected_error",
        [
            (
                [],
                "Error code 400 | Validation error: Missing required parameter 'universe'",
            ),
            (
                ["INVALID"],
                "Error code 412 | Unable to resolve all requested identifiers. Requested items: ['INVALID']",
            ),
        ],
    )
    def test_esg_full_measures_definition_object_with_invalid_params_and_get_error(
        self,
        universe,
        expected_error,
        open_platform_session,
    ):
        with pytest.raises(RDError) as error:
            full_measures.Definition(universe=universe).get_data()
        assert str(error.value) == expected_error

    @allure.title(
        "Create ESG Full Measures definition object with use mix valid and invalid data - asynchronously"
    )
    @pytest.mark.caseid("C40232229")
    @pytest.mark.parametrize(
        "universe,invalid_universe", [(["BNPP.PA", "INVALID"], "INVALID")]
    )
    async def test_esg_full_measures_definition_object_with_use_mix_valid_and_invalid_data(
        self, open_platform_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            full_measures.Definition(universe=universe),
            full_measures.Definition(universe=invalid_universe),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['INVALID']",
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['INVALID']",
        )



from enum import Enum, unique


@unique
class InterestCalculationMethod(Enum):
    DCB_30_E_360_ISMA = "Dcb_30E_360_ISMA"
    DCB_30_360 = "Dcb_30_360"
    DCB_30_360_GERMAN = "Dcb_30_360_German"
    DCB_30_360_ISDA = "Dcb_30_360_ISDA"
    DCB_30_360_US = "Dcb_30_360_US"
    DCB_30_365_BRAZIL = "Dcb_30_365_Brazil"
    DCB_30_365_GERMAN = "Dcb_30_365_German"
    DCB_30_365_ISDA = "Dcb_30_365_ISDA"
    DCB_30_ACTUAL = "Dcb_30_Actual"
    DCB_30_ACTUAL_GERMAN = "Dcb_30_Actual_German"
    DCB_30_ACTUAL_ISDA = "Dcb_30_Actual_ISDA"
    DCB_ACTUAL_LEAP_DAY_360 = "Dcb_ActualLeapDay_360"
    DCB_ACTUAL_LEAP_DAY_365 = "Dcb_ActualLeapDay_365"
    DCB_ACTUAL_360 = "Dcb_Actual_360"
    DCB_ACTUAL_364 = "Dcb_Actual_364"
    DCB_ACTUAL_365 = "Dcb_Actual_365"
    DCB_ACTUAL_36525 = "Dcb_Actual_36525"
    DCB_ACTUAL_365_L = "Dcb_Actual_365L"
    DCB_ACTUAL_365_P = "Dcb_Actual_365P"
    DCB_ACTUAL_365_CANADIAN_CONVENTION = "Dcb_Actual_365_CanadianConvention"
    DCB_ACTUAL_ACTUAL = "Dcb_Actual_Actual"
    DCB_ACTUAL_ACTUAL_AFB = "Dcb_Actual_Actual_AFB"
    DCB_ACTUAL_ACTUAL_ISDA = "Dcb_Actual_Actual_ISDA"
    DCB_CONSTANT = "Dcb_Constant"
    DCB_WORKING_DAYS_252 = "Dcb_WorkingDays_252"



from refinitiv.data._tools._converter import convert_content_data_to_df
import pandas as pd


def test_dataframe_requirements():
    content_data = {
        "headers": [
            {"name": "instrument"},
            {"name": "investorid"},
        ],
        "data": [["VOD.L", ""], ["JPN.M", None], ["IBM.N", "2004260"]],
    }
    df = convert_content_data_to_df(content_data)
    for row in df.itertuples():
        if row.instrument == "VOD.L":
            assert row.investorid == ""
        if row.instrument == "JPN.M":
            assert row.investorid is pd.NA
        if row.instrument == "IBM.N":
            assert isinstance(row.investorid, str)



import pytest

from refinitiv.data._core.session._session_type import SessionType


@pytest.fixture(scope="function", params=[SessionType.DESKTOP, SessionType.PLATFORM])
def session_type(request):
    return request.param



from ._fx_average_info import FxAverageInfo
from ._fx_barrier_definition import FxBarrierDefinition
from ._fx_binary_definition import FxBinaryDefinition
from ._fx_double_barrier_definition import FxDoubleBarrierDefinition
from ._fx_double_barrier_info import FxDoubleBarrierInfo
from ._fx_double_binary_definition import FxDoubleBinaryDefinition
from ._fx_dual_currency_definition import FxDualCurrencyDefinition
from ._fx_forward_start import FxForwardStart
from ._fx_underlying_definition import FxUnderlyingDefinition
from ._fx_definition import FxDefinition



from typing import Optional, List

from ._fx_forward_curve_definition import FxForwardCurveDefinition
from ._fx_forward_curve_parameters import FxForwardCurveParameters
from ._fx_shift_scenario import FxShiftScenario


CurveDefinition = Optional[FxForwardCurveDefinition]
CurveParameters = Optional[FxForwardCurveParameters]
ShiftScenarios = Optional[List[FxShiftScenario]]



import mock
import pytest

from refinitiv.data.content._universe_content_validator import UniverseContentValidator
from refinitiv.data.content.ownership._enums import (
    StatTypes,
    Frequency,
    SortOrder,
)
from refinitiv.data.content.ownership._ownership_data_provider import (
    OwnershipRequestFactory,
)
from refinitiv.data.delivery._data._data_provider import ParsedData


@pytest.mark.parametrize(
    ("input_value", "test_result"),
    [
        ({"universe": "test_value"}, [("universe", "test_value")]),
        (
            {"universe": "test_value", "stat_type": 1},
            [("universe", "test_value"), ("statType", "1")],
        ),
        (
            {"universe": "test_value", "stat_type": StatTypes.INVESTOR_TYPE},
            [("universe", "test_value"), ("statType", "1")],
        ),
        (
            {"universe": "test_value", "offset": "test_value"},
            [("universe", "test_value"), ("offset", "test_value")],
        ),
        (
            {"universe": "test_value", "limit": "test_value"},
            [("universe", "test_value"), ("limit", "test_value")],
        ),
        (
            {"universe": "test_value", "sort_order": "asc"},
            [("universe", "test_value"), ("sortOrder", "asc")],
        ),
        (
            {"universe": "test_value", "sort_order": SortOrder.ASCENDING},
            [("universe", "test_value"), ("sortOrder", "asc")],
        ),
        (
            {"universe": "test_value", "frequency": "M"},
            [("universe", "test_value"), ("frequency", "M")],
        ),
        (
            {"universe": "test_value", "frequency": Frequency.MONTHLY},
            [("universe", "test_value"), ("frequency", "M")],
        ),
        (
            {"universe": "test_value", "count": "test_value"},
            [("universe", "test_value"), ("count", "test_value")],
        ),
        (
            {"universe": "test_value", "start": "1Q"},
            [("universe", "test_value"), ("start", "1Q")],
        ),
        (
            {"universe": "test_value", "end": "1Q"},
            [("universe", "test_value"), ("end", "1Q")],
        ),
        (
            {"universe": "test_value", "end": None},
            [
                ("universe", "test_value"),
            ],
        ),
        (
            {"universe": "test_value", "start": "12.10.2021"},
            [("universe", "test_value"), ("start", "20211210")],
        ),
        (
            {"universe": "test_value", "end": "12/11/2020"},
            [("universe", "test_value"), ("end", "20201211")],
        ),
    ],
)
def test_ownership_request_factory_get_query_params(input_value, test_result):
    # given
    request_factory = OwnershipRequestFactory()

    # when
    result = request_factory.get_query_parameters(**input_value)

    # then
    assert result == test_result


def test_ownership_request_factory_get_query_params_with_duplicate_universe():
    # given
    request_factory = OwnershipRequestFactory()
    input_universe = ["test_value", "testt_value", "test_value"]
    expected_results = ["test_value,testt_value", "testt_value,test_value"]

    # when
    result = request_factory.get_query_parameters(universe=input_universe)

    # then
    assert dict(result)["universe"] in expected_results, result


@pytest.mark.parametrize(
    ("input_query_parameters", "input_extended_params", "test_result"),
    [
        ([], {"universe": "test_value"}, [("universe", "test_value")]),
        (
            [("universe", "test_value")],
            {"universe": "new_value"},
            [("universe", "new_value")],
        ),
        (
            [("universe", "test_value")],
            {"universe1": "new_value"},
            [("universe", "test_value"), ("universe1", "new_value")],
        ),
    ],
)
def test_ownership_request_factory_extend_query_params(
    input_query_parameters, input_extended_params, test_result
):
    # given
    request_factory = OwnershipRequestFactory()

    # when
    result = request_factory.extend_query_parameters(
        input_query_parameters, input_extended_params
    )

    # then
    assert result == test_result


@pytest.mark.parametrize(
    ("input_data", "test_result"),
    [
        ({"content_data": {}}, True),
        ({"content_data": {"error": {"code": "", "description": ""}}}, False),
    ],
)
def test_ownership_validator(input_data, test_result):
    # given
    # when
    validator = UniverseContentValidator()
    result = validator.validate(ParsedData({}, {}, **input_data))

    # then
    assert result == test_result



from refinitiv.data.content.ipa.dates_and_calendars import holidays
from refinitiv.data.content.ipa.dates_and_calendars._request_factory import (
    DatesAndCalendarsResponseFactory,
)
from refinitiv.data.content.ipa.dates_and_calendars.holidays._holidays_data_provider import (
    HolidaysData,
)
from refinitiv.data.delivery._data._data_provider import ParsedData

MOCKED_HOLIDAYS_DATA = ParsedData(
    {"http_status_code": 200, "http_reason": "OK"},
    {},
    [
        {
            "tag": "my request",
            "holidays": [
                {
                    "date": "2018-12-31",
                    "names": [
                        {
                            "name": "New Year's Bridge Holiday '18",
                            "calendars": ["UKR"],
                            "countries": ["UKR"],
                        }
                    ],
                    "calendars": ["UKR"],
                    "countries": ["UKR"],
                },
                {
                    "date": "2019-01-01",
                    "names": [
                        {
                            "name": "New Year's Day (day 1)",
                            "calendars": ["UKR"],
                            "countries": ["UKR"],
                        },
                        {
                            "name": "New Year's Day",
                            "calendars": ["FRA", "EMU"],
                            "countries": ["FRA", ""],
                        },
                    ],
                    "calendars": ["UKR", "FRA", "EMU"],
                    "countries": ["UKR", "FRA", ""],
                },
                {
                    "date": "2019-01-02",
                    "names": [
                        {
                            "name": "New Year's Day (Day 2) '19",
                            "calendars": ["UKR"],
                            "countries": ["UKR"],
                        }
                    ],
                    "calendars": ["UKR"],
                    "countries": ["UKR"],
                },
            ],
        }
    ],
)

MOCKED_DATA_ERROR = ParsedData(
    {"http_status_code": 200, "http_reason": "OK"},
    {},
    [
        {
            "tag": "my request",
            "error": {
                "status": "Error",
                "message": "Business Calendar: calendar not found.",
                "code": "IPA_library.ErrorBusinessCalendar_CalendarNotFound",
            },
        }
    ],
    ["IPA_library.ErrorBusinessCalendar_CalendarNotFound"],
    ["Business Calendar: calendar not found."],
)


def test_holidays():
    definition = holidays.Definition(
        tag="my request",
        start_date="2018-12-31",
        end_date="2019-01-03",
        calendars=["UKR", "FRA"],
        currencies=["EUR"],
        holiday_outputs=[
            holidays.HolidayOutputs.DATE,
            holidays.HolidayOutputs.NAMES,
            holidays.HolidayOutputs.CALENDARS,
            holidays.HolidayOutputs.COUNTRIES,
        ],
    )

    assert definition.request_item.tag == "my request"
    assert definition.request_item.start_date == "2018-12-31"
    assert definition.request_item.end_date == "2019-01-03"
    assert definition.request_item.calendars == ["UKR", "FRA"]
    assert definition.request_item.currencies == ["EUR"]
    assert definition.request_item.holiday_outputs == [
        holidays.HolidayOutputs.DATE,
        holidays.HolidayOutputs.NAMES,
        holidays.HolidayOutputs.CALENDARS,
        holidays.HolidayOutputs.COUNTRIES,
    ]


def test_holidays_create_success():
    response_factory = DatesAndCalendarsResponseFactory(data_class=HolidaysData)
    response = response_factory.create_success(MOCKED_HOLIDAYS_DATA)
    assert not response.data.df.empty
    assert response.data.raw == MOCKED_HOLIDAYS_DATA.content_data
    assert response.data.holidays[1].countries == ["UKR", "FRA", ""]
    assert response.data.holidays[0].tag == "my request"


def test_holidays_create_fail():
    response_factory = DatesAndCalendarsResponseFactory(data_class=HolidaysData)
    response = response_factory.create_fail(MOCKED_DATA_ERROR)
    assert response.data.raw == MOCKED_DATA_ERROR.content_data
    assert len(response.errors)



from typing import Any, Callable, Dict, TYPE_CHECKING

from ..delivery._data._endpoint_data import EndpointData

if TYPE_CHECKING:
    import pandas as pd


class Data(EndpointData):
    _dataframe = None

    def __init__(
        self,
        raw: Any,
        dfbuilder: Callable[[Any, Dict[str, Any]], "pd.DataFrame"],
        **kwargs,
    ):
        EndpointData.__init__(self, raw, **kwargs)
        self._dfbuilder = dfbuilder

    @property
    def df(self):
        if self._dataframe is None:
            self._dataframe = self._dfbuilder(self.raw, **self._kwargs)

        return self._dataframe



from datetime import timedelta, datetime

import allure
import pytest

from refinitiv.data.content import historical_pricing
from refinitiv.data.content.historical_pricing import MarketSession
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason, Intervals
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_non_empty_response_data,
    check_response_status,
    check_response_is_success,
    check_response_values,
    check_universe_in_response,
    is_expected_param_in_request_url,
    check_index_column_contains_dates,
    check_universe_order_in_df,
    check_extended_params_were_sent_in_request,
)


@allure.suite("Historical Pricing Summaries")
@allure.feature("Historical Pricing Summaries")
@allure.severity(allure.severity_level.CRITICAL)
class TestHistoricalPricingSummaries:
    @allure.title(
        "Create HistoricalPricing Summaries Definition object using IntradayInterval and extended params"
    )
    @pytest.mark.parametrize(
        "universe,intervals,sessions,extended_params",
        [
            (
                "GBP=",
                historical_pricing.Intervals.ONE_MINUTE,
                [MarketSession.PRE, MarketSession.NORMAL, MarketSession.POST],
                None,
            ),
            (
                ["VOD.L", "GOOG.O", "LSEG.L"],
                "P1M",
                None,
                {"sessions": "pre,normal,post"},
            ),
        ],
    )
    @pytest.mark.caseid("33303593")
    def test_create_historical_pricing_summaries_object_with_universe_intraday_interval_and_extended_params(
        self,
        open_desktop_session,
        universe,
        intervals,
        sessions,
        extended_params,
    ):
        response = historical_pricing.summaries.Definition(
            universe,
            interval=intervals,
            count=200,
            sessions=sessions,
            extended_params=extended_params,
        ).get_data()
        check_universe_in_response(response, universe)
        check_index_column_contains_dates(response)
        check_universe_order_in_df(response, universe)
        check_extended_params_were_sent_in_request(response, extended_params)

    @allure.title(
        "Create HistoricalPricing Summaries Definition object using a string/list of strings to define universe and InterdayInterval"
    )
    @pytest.mark.caseid("33303593")
    @pytest.mark.parametrize(
        "universe,intervals",
        [("VOD.L", historical_pricing.Intervals.DAILY), (["VOD.L", "GOOG.O"], None)],
    )
    def test_create_historical_pricing_summaries_object_with_universe_interday_interval(
        self, open_desktop_session, universe, intervals
    ):
        response = historical_pricing.summaries.Definition(
            universe, interval=intervals
        ).get_data()

        check_universe_in_response(response, universe)
        check_index_column_contains_dates(response)

    @allure.title(
        "Create HistoricalPricing Summaries Definition object using some valid optional parameters"
    )
    @pytest.mark.caseid("33303594")
    @pytest.mark.parametrize(
        "universe, fields, start_date, end_date",
        [
            ("EUR=", ["DATE", "MKT_OPEN", "BID", "ASK"], timedelta(-3), timedelta(0)),
            ("EUR=", "BID", timedelta(hours=-100), datetime.now()),
            ("IBM", None, "2021-12-30T00:10:00", "2022-01-05T05:10:00"),
        ],
        ids=["timedelta", "datetime", "string"],
    )
    def test_create_historical_pricing_summaries_object_with_valid_optional_parameters(
        self, open_desktop_session, universe, fields, start_date, end_date
    ):
        response = historical_pricing.summaries.Definition(
            universe,
            interval=historical_pricing.Intervals.DAILY,
            adjustments=historical_pricing.Adjustments.CCH,
            count=5,
            fields=fields,
            start=start_date,
            end=end_date,
        ).get_data()
        is_expected_param_in_request_url(response, ["start", "end"])
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_values(
            response,
            expected_universe_ric=universe,
            expected_interval=Intervals.DAILY,
            expected_summary_timestamp_label="endPeriod",
        )

    @allure.title(
        "Create HistoricalPricing Summaries Definition object using invalid universe"
    )
    @pytest.mark.caseid("33303595")
    @pytest.mark.parametrize("universe", ["VOD.invalid"])
    def test_create_historical_pricing_summaries_object_with_invalid_universe(
        self, open_desktop_session, universe
    ):
        with pytest.raises(RDError) as error:
            historical_pricing.summaries.Definition(universe).get_data()
        assert (
            str(error.value)
            == "Error code 1 | No data to return, please check errors: ERROR: No successful "
            "response.\n"
            "(TS.Interday.UserRequestError.70005, The universe is not found)"
        )

    @allure.title(
        "Create HistoricalPricing Summaries Definition object using invalid Interval"
    )
    @pytest.mark.caseid("33303596")
    @pytest.mark.parametrize("universe", ["VOD.L"])
    def test_create_historical_pricing_summaries_object_using_invalid_interval(
        self, open_desktop_session, universe
    ):
        """
        we have different with TS lib, when creating a request using invalid Interval in TS lib expected status code 400
        in python lib we receive AttributeError "interval must be in ['PT1M', 'PT5M', 'PT10M', 'PT30M', 'PT60M', 'PT1H',
        'P1D', 'P7D', 'P1W', 'P1M', 'P3M', 'P12M', 'P1Y']"
        this test case only catch AttributeError, if AttributeError not rise we assert on http status code
        """
        with pytest.raises(AttributeError) as error:
            historical_pricing.summaries.Definition(
                universe, interval="INVALID"
            ).get_data()
        assert (
            str(error.value)
            == "Value 'INVALID' must be in ['PT1M', 'PT5M', 'PT10M', 'PT30M', 'PT60M', 'PT1H', "
            "'P1D', 'P7D', 'P1W', 'P1M', 'P3M', 'P12M', 'P1Y']"
        )

    @allure.title(
        "Create HistoricalPricing Summaries Definition object when session is not open"
    )
    @pytest.mark.parametrize("universe", ["VOD.L"])
    def test_create_historical_pricing_summaries_object_with_session_not_opened(
        self, universe, open_desktop_session
    ):
        session = open_desktop_session
        session.close()
        historical_pricing_definition = historical_pricing.summaries.Definition(
            universe
        )
        with pytest.raises(ValueError) as error:
            historical_pricing_definition.get_data()
        assert str(error.value) == "Session is not opened. Can't send any request"

    @allure.title(
        "Create Historical Pricing Summaries get data async with valid and invalid universe"
    )
    @pytest.mark.parametrize(
        "universe,invalid_universe", [(["VOD.L", "INVALID"], "INVALID")]
    )
    @pytest.mark.caseid("39746370")
    async def test_get_data_async(
        self, open_desktop_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            historical_pricing.summaries.Definition(universe),
            historical_pricing.summaries.Definition(invalid_universe),
        )
        check_response_is_success(valid_response)
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="TS.Interday.UserRequestError.70005",
            expected_error_message="The universe is not found.. Requested ric: INVALID. Requested fields: None",
        )
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="TS.Interday.UserRequestError.70005",
            expected_error_message="The universe is not found.. Requested ric: INVALID. Requested fields: None",
        )



from typing import Union

from ._session import Session
from ._session_cxn_type import SessionCxnType
from .connection import (
    RefinitivDataConnection,
    RefinitivDataAndDeployedConnection,
    DesktopConnection,
    DeployedConnection,
)

cxn_class_by_type = {
    SessionCxnType.DEPLOYED: DeployedConnection,
    SessionCxnType.REFINITIV_DATA: RefinitivDataConnection,
    SessionCxnType.REFINITIV_DATA_AND_DEPLOYED: RefinitivDataAndDeployedConnection,
    SessionCxnType.DESKTOP: DesktopConnection,
}

SessionConnection = Union[
    RefinitivDataConnection,
    RefinitivDataAndDeployedConnection,
    DesktopConnection,
    DeployedConnection,
]

PlatformConnection = Union[
    RefinitivDataConnection, RefinitivDataAndDeployedConnection, DeployedConnection
]


def get_session_cxn(
    session_cxn_type: SessionCxnType, session: "Session"
) -> SessionConnection:
    cxn_class = cxn_class_by_type.get(session_cxn_type)

    if not cxn_class:
        raise ValueError(
            f"Can't find cxn_class by session_cxn_type: {session_cxn_type}"
        )

    cxn = cxn_class(session)
    return cxn



from typing import List

from ._adc_context import ADCContext
from ._context import UDFMixin


class ADCUDFContext(UDFMixin, ADCContext):
    def prepare_headers(self, raw: dict, *args) -> List[List[str]]:
        """
        Prepare headers to pass them to df_builder.

        Args:
            raw (dict): adc raw data from adc response.

        Returns:
            List with inserted list of dicts that describe each header in understandable
            by df builder format.
        """
        fields = self.fields.raw

        field_to_idx = {field.upper(): fields.index(field) + 2 for field in fields}
        headers = [{"displayName": "Instrument"}, {"displayName": "Date"}]

        for header in raw["headers"][0]:
            if len(header) > 1:
                field_upper = header["field"].upper()

                if field_upper not in field_to_idx:
                    continue

                headers.insert(field_to_idx[field_upper], header)
                field_to_idx.pop(field_upper)

        for field, index in field_to_idx.items():
            headers.insert(index, {"displayName": field, "field": field})

        return [headers]

    def get_cols(self, headers, *args) -> List[str]:
        return [col["name"] for col in headers if col["name"] != "Instrument"]

    def is_raw_headers_in_fields(self, raw, fields) -> bool:
        fields = {field.casefold() for field in fields}
        headers = set()

        for header in raw["headers"][0]:
            if header.get("field"):
                headers.add(header.get("field").casefold())

        return headers.issubset(fields)



import pytest

from refinitiv.data.content.news import Urgency
from refinitiv.data.content.news._data_classes import HeadlineRDP, Story
from refinitiv.data.content.news._news_data_provider import (
    NewsStoryRDPData,
    NewsStoryUDFData,
)

HEADLINE_RAW_JSON = {
    "storyId": "urn:newsml:reuters.com:20210713:nBw2Wmkgya:1",
    "newsItem": {
        "_version": 1,
        "contentMeta": {
            "audience": [{"_qcode": "NP:BSW"}, {"_qcode": "NP:CNR"}],
            "creator": [{"_qcode": "NS:BSW", "_role": "sRole:source"}],
            "infoSource": [
                {"_qcode": "NS:BSW", "_role": "sRole:source"},
                {"_qcode": "NS:BSW", "_role": "sRole:origProv"},
            ],
            "language": [{"_tag": "en"}],
            "subject": [
                {"_qcode": "B:255"},
                {"_qcode": "M:27I"},
                {"_qcode": "B:198"},
                {"_qcode": "G:4"},
                {"_qcode": "M:1WK"},
                {"_qcode": "B:162"},
                {"_qcode": "B:279"},
                {"_qcode": "M:1TG"},
                {"_qcode": "B:69"},
                {"_qcode": "M:Z"},
                {"_qcode": "B:291"},
                {"_qcode": "G:9"},
                {"_qcode": "M:E7"},
                {"_qcode": "M:2CQ"},
                {"_qcode": "B:278"},
                {"_qcode": "M:1WJ"},
                {"_qcode": "M:1TH"},
                {"_qcode": "M:1WN"},
                {"_qcode": "B:161"},
                {"_qcode": "B:282"},
                {"_qcode": "E:1"},
                {"_qcode": "M:2CX"},
                {"_qcode": "M:1T4"},
                {"_qcode": "G:6J"},
                {"_qcode": "M:2CP"},
                {"_qcode": "M:2D1"},
                {"_qcode": "M:2CS"},
                {"_qcode": "M:NL"},
                {"_qcode": "R:GOOGL.O"},
                {"_qcode": "B:285"},
                {"_qcode": "M:23G"},
                {"_qcode": "B:172"},
                {"_qcode": "M:1ZD"},
                {"_qcode": "B:257"},
                {"_qcode": "M:32"},
                {"_qcode": "B:75"},
                {"_qcode": "B:87"},
                {"_qcode": "G:50"},
                {"_qcode": "M:V"},
                {"_qcode": "B:256"},
                {"_qcode": "M:1QD"},
                {"_qcode": "M:2CR"},
                {"_qcode": "M:1WL"},
                {"_qcode": "B:173"},
                {"_qcode": "M:1ZN"},
                {"_qcode": "R:005930.KS"},
                {"_qcode": "B:86"},
                {"_qcode": "M:1P2"},
                {"_qcode": "E:6Q"},
                {"_qcode": "E:40"},
                {"_qcode": "B:78"},
                {"_qcode": "B:290"},
                {"_qcode": "B:1805"},
                {"_qcode": "M:2CN"},
                {"_qcode": "P:4295882451"},
                {"_qcode": "P:5046736305"},
                {"_qcode": "P:5030853586"},
                {"_qcode": "P:4298538029"},
                {"_qcode": "P:4295899948"},
            ],
            "urgency": {"$": 3},
        },
        "itemMeta": {
            "firstCreated": {"$": "2021-07-13T14:00:00.749Z"},
            "versionCreated": {"$": "2021-07-13T14:00:00.749Z"},
            "title": [
                {
                    "$": "Backtracks Releases New Integration with Google "
                    "Cloud for Podcast Audience and Engagement Data"
                }
            ],
        },
    },
}

STORY_RAW_JSON = {
    "newsItem": {
        "contentMeta": {
            "urgency": {"$": 3},
            "subject": [
                {
                    "_confidence": 100,
                    "_creator": "rftResRef:sys26",
                    "_how": "howextr:tool",
                    "_id": "S1",
                    "_qcode": "B:161",
                    "_why": "why:inferred",
                    "related": [
                        {
                            "_creator": "rftResRef:sys24",
                            "_how": "howextr:tool",
                            "_id": "grp0",
                            "_qcode": "hmlInd:high",
                            "_rel": "extCptRel:hasRelevanceGroup",
                            "_why": "why:inferred",
                        }
                    ],
                }
            ],
            "creator": [{"_qcode": "mocked_qcode"}],
            "infoSource": "mocked_infoSource",
            "language": "mocked_language",
        },
        "itemMeta": {
            "versionCreated": {"$": "2021-05-24T04:13:02Z"},
            "firstCreated": {"$": "2021-05-24T04:13:02Z"},
            "title": [{"$": "mocked_title"}],
        },
    }
}


@pytest.mark.parametrize("news_story_class", [NewsStoryRDPData, NewsStoryUDFData])
def test_story_response(news_story_class):
    raw_json = {"mocked": "json"}
    result = news_story_class(raw_json)

    assert result.raw == raw_json


def test_headlines_create_values():
    headline = HeadlineRDP.from_dict(HEADLINE_RAW_JSON)
    assert headline.creator == "NS:BSW"
    assert headline.language == [{"_tag": "en"}]
    assert headline.source == [
        {"_qcode": "NS:BSW", "_role": "sRole:source"},
        {"_qcode": "NS:BSW", "_role": "sRole:origProv"},
    ]
    assert headline.story_id == "urn:newsml:reuters.com:20210713:nBw2Wmkgya:1"
    assert (
        headline.title == "Backtracks Releases New Integration with Google"
        " Cloud for Podcast Audience and Engagement Data"
    )
    assert headline.urgency == Urgency.Regular


def test_story_create_values():
    story = Story.from_dict(STORY_RAW_JSON)
    assert story.creator == "mocked_qcode"
    assert story.item_codes == ["B:161"]
    assert story.language == "mocked_language"
    assert story.source == "mocked_infoSource"
    assert story.title == "mocked_title"
    assert story.urgency == Urgency.Regular



from typing import Union, List

from ._hp_context import HPContext


class HPAndCustInstContext(HPContext):
    @property
    def dfbuilder(self) -> "DFBuilder":
        pass

    @property
    def date_name(self) -> str:
        pass

    def prepare_headers(self, raw: Union[list, dict], *args) -> List[List[str]]:
        pass

    @property
    def can_get_data(self):
        return super().can_get_data or self.universe.cust_inst



from typing import TYPE_CHECKING

from ._data_provider import TopNewsData
from ...._content_type import ContentType
from ....delivery._data._data_provider_layer import DataProviderLayer
from ....delivery._data._response import BaseResponse

if TYPE_CHECKING:
    from ...._types import OptInt


class Definition(DataProviderLayer[BaseResponse[TopNewsData]]):
    """
    This class describes parameters to retrieve data for top news headlines.

    Parameters
    ----------
    top_news_id: str
        The id of the package

    revision_id: int, optional
        The package known version

     Examples
    --------
    >>> from refinitiv.data.content import news
    >>> definition = news.top_news.Definition("top_news_id")
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        top_news_id: str,
        revision_id: "OptInt" = None,
    ):
        super().__init__(
            data_type=ContentType.NEWS_TOP_NEWS_HEADLINES,
            top_news_id=top_news_id,
            revision_id=revision_id,
        )



from typing import Optional, TYPE_CHECKING

from ._override_bid_ask_fields import OverrideBidAskFields
from ...._object_definition import ObjectDefinition

if TYPE_CHECKING:
    from ......_types import OptStr


class OverrideBidAsk(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    instrument_code : str, optional
        Reuters instrument code (ric)
    fields : OverrideBidAskFields, optional

    date : str, optional
        Overridden date
    """

    def __init__(
        self,
        instrument_code: "OptStr" = None,
        fields: Optional[OverrideBidAskFields] = None,
        date: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.instrument_code = instrument_code
        self.fields = fields
        self.date = date

    @property
    def fields(self):
        """
        :return: object OverrideBidAskFields
        """
        return self._get_object_parameter(OverrideBidAskFields, "fields")

    @fields.setter
    def fields(self, value):
        self._set_object_parameter(OverrideBidAskFields, "fields", value)

    @property
    def date(self):
        """
        Overridden date
        :return: str
        """
        return self._get_parameter("date")

    @date.setter
    def date(self, value):
        self._set_parameter("date", value)

    @property
    def instrument_code(self):
        """
        Reuters instrument code (ric)
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)
